"""This module defines actions (functions) that operate on layers and its submenus.

Among other potential uses, these will populate the menu when you right-click
on a layer in the LayerList.

The Actions in LAYER_ACTIONS are registered with the application when it is
created in `_app_model._app`.  Modifying this list at runtime will have no
effect.  Use `app.register_action` to register new actions at runtime.
"""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

from app_model.types import Action, SubmenuItem

from napari._app_model.constants import MenuGroup, MenuId
from napari._app_model.context import LayerListSelectionContextKeys as LLSCK
from napari.layers import _layer_actions
from napari.utils.translations import trans

if TYPE_CHECKING:
    from app_model.types import MenuRuleDict

# Layer submenus
LAYERLIST_CONTEXT_SUBMENUS = [
    (
        MenuId.LAYERLIST_CONTEXT,
        SubmenuItem(
            submenu=MenuId.LAYERS_CONTEXT_CONVERT_DTYPE,
            title=trans._('Convert data type'),
            group=MenuGroup.LAYERLIST_CONTEXT.CONVERSION,
            order=None,
            enablement=LLSCK.all_selected_layers_labels,
        ),
    ),
    (
        MenuId.LAYERLIST_CONTEXT,
        SubmenuItem(
            submenu=MenuId.LAYERS_CONTEXT_PROJECT,
            title=trans._('Projections'),
            group=MenuGroup.LAYERLIST_CONTEXT.SPLIT_MERGE,
            order=None,
            enablement=LLSCK.active_layer_is_image_3d,
        ),
    ),
    (
        MenuId.LAYERLIST_CONTEXT,
        SubmenuItem(
            submenu=MenuId.LAYERS_CONTEXT_COPY_SPATIAL,
            title=trans._('Copy scale and transforms'),
            group=MenuGroup.LAYERLIST_CONTEXT.COPY_SPATIAL,
            order=None,
            enablement=(LLSCK.num_selected_layers == 1),
        ),
    ),
]

# The following dicts define groups to which menu items in the layer list context menu can belong
# see https://app-model.readthedocs.io/en/latest/types/#app_model.types.MenuRule for details
LAYERCTX_SPLITMERGE: MenuRuleDict = {
    'id': MenuId.LAYERLIST_CONTEXT,
    'group': MenuGroup.LAYERLIST_CONTEXT.SPLIT_MERGE,
}
LAYERCTX_CONVERSION: MenuRuleDict = {
    'id': MenuId.LAYERLIST_CONTEXT,
    'group': MenuGroup.LAYERLIST_CONTEXT.CONVERSION,
}
LAYERCTX_LINK: MenuRuleDict = {
    'id': MenuId.LAYERLIST_CONTEXT,
    'group': MenuGroup.LAYERLIST_CONTEXT.LINK,
}

# Statically defined Layer actions.
# modifying this list at runtime has no effect.
LAYERLIST_CONTEXT_ACTIONS: list[Action] = [
    Action(
        id='napari.layer.duplicate',
        title=trans._('Duplicate Layer'),
        callback=_layer_actions._duplicate_layer,
        menus=[LAYERCTX_SPLITMERGE],
    ),
    Action(
        id='napari.layer.split_stack',
        title=trans._('Split Stack'),
        callback=_layer_actions._split_stack,
        menus=[{**LAYERCTX_SPLITMERGE, 'when': ~LLSCK.active_layer_is_rgb}],
        enablement=LLSCK.active_layer_is_image_3d,
    ),
    Action(
        id='napari.layer.split_rgb',
        title=trans._('Split RGB'),
        callback=_layer_actions._split_rgb,
        menus=[{**LAYERCTX_SPLITMERGE, 'when': LLSCK.active_layer_is_rgb}],
        enablement=LLSCK.active_layer_is_rgb,
    ),
    Action(
        id='napari.layer.convert_to_labels',
        title=trans._('Convert to Labels'),
        callback=_layer_actions._convert_to_labels,
        enablement=(
            (
                (LLSCK.num_selected_image_layers >= 1)
                | (LLSCK.num_selected_shapes_layers >= 1)
            )
            & LLSCK.all_selected_layers_same_type
            & ~LLSCK.selected_empty_shapes_layer
        ),
        menus=[LAYERCTX_CONVERSION],
    ),
    Action(
        id='napari.layer.convert_to_image',
        title=trans._('Convert to Image'),
        callback=_layer_actions._convert_to_image,
        enablement=(
            (LLSCK.num_selected_labels_layers >= 1)
            & LLSCK.all_selected_layers_same_type
        ),
        menus=[LAYERCTX_CONVERSION],
    ),
    Action(
        id='napari.layer.merge_stack',
        title=trans._('Merge to Stack'),
        callback=_layer_actions._merge_stack,
        enablement=(
            (LLSCK.num_selected_layers > 1)
            & (LLSCK.num_selected_image_layers == LLSCK.num_selected_layers)
            & LLSCK.all_selected_layers_same_shape
        ),
        menus=[LAYERCTX_SPLITMERGE],
    ),
    Action(
        id='napari.layer.toggle_visibility',
        title=trans._('Toggle visibility'),
        callback=_layer_actions._toggle_visibility,
        menus=[
            {
                'id': MenuId.LAYERLIST_CONTEXT,
                'group': MenuGroup.NAVIGATION,
            }
        ],
    ),
    Action(
        id='napari.layer.link_selected_layers',
        title=trans._('Link Layers'),
        callback=_layer_actions._link_selected_layers,
        enablement=(
            (LLSCK.num_selected_layers > 1) & ~LLSCK.num_selected_layers_linked
        ),
        menus=[{**LAYERCTX_LINK, 'when': ~LLSCK.num_selected_layers_linked}],
    ),
    Action(
        id='napari.layer.unlink_selected_layers',
        title=trans._('Unlink Layers'),
        callback=_layer_actions._unlink_selected_layers,
        enablement=LLSCK.num_selected_layers_linked,
        menus=[{**LAYERCTX_LINK, 'when': LLSCK.num_selected_layers_linked}],
    ),
    Action(
        id='napari.layer.select_linked_layers',
        title=trans._('Select Linked Layers'),
        callback=_layer_actions._select_linked_layers,
        enablement=LLSCK.num_unselected_linked_layers,
        menus=[LAYERCTX_LINK],
    ),
    Action(
        id='napari.layer.show_selected',
        title=trans._('Show All Selected Layers'),
        callback=_layer_actions._show_selected,
        menus=[
            {
                'id': MenuId.LAYERLIST_CONTEXT,
                'group': MenuGroup.NAVIGATION,
            }
        ],
    ),
    Action(
        id='napari.layer.hide_selected',
        title=trans._('Hide All Selected Layers'),
        callback=_layer_actions._hide_selected,
        menus=[
            {
                'id': MenuId.LAYERLIST_CONTEXT,
                'group': MenuGroup.NAVIGATION,
            }
        ],
    ),
    Action(
        id='napari.layer.show_unselected',
        title=trans._('Show All Unselected Layers'),
        callback=_layer_actions._show_unselected,
        menus=[
            {
                'id': MenuId.LAYERLIST_CONTEXT,
                'group': MenuGroup.NAVIGATION,
            }
        ],
    ),
    Action(
        id='napari.layer.hide_unselected',
        title=trans._('Hide All Unselected Layers'),
        callback=_layer_actions._hide_unselected,
        menus=[
            {
                'id': MenuId.LAYERLIST_CONTEXT,
                'group': MenuGroup.NAVIGATION,
            }
        ],
    ),
]

for _dtype in (
    'int8',
    'int16',
    'int32',
    'int64',
    'uint8',
    'uint16',
    'uint32',
    'uint64',
):
    LAYERLIST_CONTEXT_ACTIONS.append(
        Action(
            id=f'napari.layer.convert_to_{_dtype}',
            title=trans._('Convert to {dtype}', dtype=_dtype),
            callback=partial(_layer_actions._convert_dtype, mode=_dtype),
            enablement=(
                LLSCK.all_selected_layers_labels
                & (LLSCK.active_layer_dtype != _dtype)
            ),
            menus=[{'id': MenuId.LAYERS_CONTEXT_CONVERT_DTYPE}],
        )
    )

for mode in ('max', 'min', 'std', 'sum', 'mean', 'median'):
    LAYERLIST_CONTEXT_ACTIONS.append(
        Action(
            id=f'napari.layer.project_{mode}',
            title=trans._('{mode} projection', mode=mode),
            callback=partial(_layer_actions._project, mode=mode),
            enablement=LLSCK.active_layer_is_image_3d,
            menus=[{'id': MenuId.LAYERS_CONTEXT_PROJECT}],
        )
    )



import numpy as np
import numpy.testing as npt
import pint
import pytest
from scipy.stats import special_ortho_group

from napari.utils.transforms import Affine, CompositeAffine, ScaleTranslate

transform_types = [Affine, CompositeAffine, ScaleTranslate]

affine_type = [Affine, CompositeAffine]

REG = pint.get_application_registry()
PIXEL = REG.pixel


@pytest.mark.parametrize('Transform', transform_types)
def test_scale_translate(Transform):
    coord = [10, 13]
    transform = Transform(scale=[2, 3], translate=[8, -5], name='st')
    assert transform._is_diagonal
    new_coord = transform(coord)
    target_coord = [2 * 10 + 8, 3 * 13 - 5]
    assert transform.name == 'st'
    npt.assert_allclose(new_coord, target_coord)


@pytest.mark.parametrize('Transform', [Affine, CompositeAffine])
def test_affine_is_diagonal(Transform):
    transform = Transform(scale=[2, 3], translate=[8, -5], name='st')
    assert transform._is_diagonal
    transform.rotate = 5.0
    assert not transform._is_diagonal
    # Rotation back to 0.0 will result in tiny non-zero off-diagonal values.
    # _is_diagonal assumes values below 1e-8 are equivalent to 0.
    transform.rotate = 0.0
    assert transform._is_diagonal


def test_diagonal_scale_setter():
    diag_transform = Affine(scale=[2, 3], name='st')
    assert diag_transform._is_diagonal
    diag_transform.scale = [1]
    npt.assert_allclose(diag_transform.scale, [1.0, 1.0])


@pytest.mark.parametrize('Transform', transform_types)
def test_scale_translate_broadcast_scale(Transform):
    coord = [1, 10, 13]
    transform = Transform(scale=[4, 2, 3], translate=[8, -5], name='st')
    new_coord = transform(coord)
    target_coord = [4, 2 * 10 + 8, 3 * 13 - 5]
    assert transform.name == 'st'
    npt.assert_allclose(transform.scale, [4, 2, 3])
    npt.assert_allclose(transform.translate, [0, 8, -5])
    npt.assert_allclose(new_coord, target_coord)


@pytest.mark.parametrize('Transform', transform_types)
def test_scale_translate_broadcast_translate(Transform):
    coord = [1, 10, 13]
    transform = Transform(scale=[2, 3], translate=[5, 8, -5], name='st')
    new_coord = transform(coord)
    target_coord = [6, 2 * 10 + 8, 3 * 13 - 5]
    assert transform.name == 'st'
    npt.assert_allclose(transform.scale, [1, 2, 3])
    npt.assert_allclose(transform.translate, [5, 8, -5])
    npt.assert_allclose(new_coord, target_coord)


@pytest.mark.parametrize('Transform', transform_types)
def test_scale_translate_inverse(Transform):
    coord = [10, 13]
    transform = Transform(scale=[2, 3], translate=[8, -5])
    new_coord = transform(coord)
    target_coord = [2 * 10 + 8, 3 * 13 - 5]
    npt.assert_allclose(new_coord, target_coord)

    inverted_new_coord = transform.inverse(new_coord)
    npt.assert_allclose(inverted_new_coord, coord)


@pytest.mark.parametrize('Transform', transform_types)
def test_scale_translate_compose(Transform):
    coord = [10, 13]
    transform_a = Transform(scale=[2, 3], translate=[8, -5])
    transform_b = Transform(scale=[0.3, 1.4], translate=[-2.2, 3])
    transform_c = transform_b.compose(transform_a)

    new_coord_1 = transform_c(coord)
    new_coord_2 = transform_b(transform_a(coord))
    npt.assert_allclose(new_coord_1, new_coord_2)


@pytest.mark.parametrize('Transform', transform_types)
def test_scale_translate_slice(Transform):
    transform_a = Transform(scale=[2, 3], translate=[8, -5])
    transform_b = Transform(scale=[2, 1, 3], translate=[8, 3, -5], name='st')
    npt.assert_allclose(transform_b.set_slice([0, 2]).scale, transform_a.scale)
    npt.assert_allclose(
        transform_b.set_slice([0, 2]).translate, transform_a.translate
    )
    assert transform_b.set_slice([0, 2]).name == 'st'


@pytest.mark.parametrize('Transform', transform_types)
def test_scale_translate_expand_dims(Transform):
    transform_a = Transform(scale=[2, 3], translate=[8, -5], name='st')
    transform_b = Transform(scale=[2, 1, 3], translate=[8, 0, -5])
    npt.assert_allclose(transform_a.expand_dims([1]).scale, transform_b.scale)
    npt.assert_allclose(
        transform_a.expand_dims([1]).translate, transform_b.translate
    )
    assert transform_a.expand_dims([1]).name == 'st'


@pytest.mark.parametrize('Transform', transform_types)
def test_scale_translate_identity_default(Transform):
    coord = [10, 13]
    transform = Transform()
    new_coord = transform(coord)
    npt.assert_allclose(new_coord, coord)


def test_affine_properties():
    transform = Affine(scale=[2, 3], translate=[8, -5], rotate=90, shear=[1])
    npt.assert_allclose(transform.translate, [8, -5])
    npt.assert_allclose(transform.scale, [2, 3])
    npt.assert_almost_equal(transform.rotate, [[0, -1], [1, 0]])
    npt.assert_almost_equal(transform.shear, [1])


def test_affine_properties_setters():
    transform = Affine()
    transform.translate = [8, -5]
    npt.assert_allclose(transform.translate, [8, -5])
    transform.scale = [2, 3]
    npt.assert_allclose(transform.scale, [2, 3])
    transform.rotate = 90
    npt.assert_almost_equal(transform.rotate, [[0, -1], [1, 0]])
    transform.shear = [1]
    npt.assert_almost_equal(transform.shear, [1])


def test_rotate():
    coord = [10, 13]
    transform = Affine(rotate=90)
    new_coord = transform(coord)
    # As rotate by 90 degrees, can use [-y, x]
    target_coord = [-coord[1], coord[0]]
    npt.assert_allclose(new_coord, target_coord)


def test_scale_translate_rotate():
    coord = [10, 13]
    transform = Affine(scale=[2, 3], translate=[8, -5], rotate=90)
    new_coord = transform(coord)
    post_scale = np.multiply(coord, [2, 3])
    # As rotate by 90 degrees, can use [-y, x]
    post_rotate = [-post_scale[1], post_scale[0]]
    target_coord = np.add(post_rotate, [8, -5])
    npt.assert_allclose(new_coord, target_coord)


def test_scale_translate_rotate_inverse():
    coord = [10, 13]
    transform = Affine(scale=[2, 3], translate=[8, -5], rotate=90)
    new_coord = transform(coord)
    post_scale = np.multiply(coord, [2, 3])
    # As rotate by 90 degrees, can use [-y, x]
    post_rotate = [-post_scale[1], post_scale[0]]
    target_coord = np.add(post_rotate, [8, -5])
    npt.assert_allclose(new_coord, target_coord)

    inverted_new_coord = transform.inverse(new_coord)
    npt.assert_allclose(inverted_new_coord, coord)


def test_scale_translate_rotate_compose():
    coord = [10, 13]
    transform_a = Affine(scale=[2, 3], translate=[8, -5], rotate=25)
    transform_b = Affine(scale=[0.3, 1.4], translate=[-2.2, 3], rotate=65)
    transform_c = transform_b.compose(transform_a)

    new_coord_1 = transform_c(coord)
    new_coord_2 = transform_b(transform_a(coord))
    npt.assert_allclose(new_coord_1, new_coord_2)


def test_scale_translate_rotate_shear_compose():
    coord = [10, 13]
    transform_a = Affine(scale=[2, 3], translate=[8, -5], rotate=25, shear=[1])
    transform_b = Affine(
        scale=[0.3, 1.4],
        translate=[-2.2, 3],
        rotate=65,
        shear=[-0.5],
    )
    transform_c = transform_b.compose(transform_a)

    new_coord_1 = transform_c(coord)
    new_coord_2 = transform_b(transform_a(coord))
    npt.assert_allclose(new_coord_1, new_coord_2)


@pytest.mark.parametrize('dimensionality', [2, 3])
def test_affine_matrix(dimensionality):
    np.random.seed(0)
    N = dimensionality
    A = np.eye(N + 1)
    A[:-1, :-1] = np.random.random((N, N))
    A[:-1, -1] = np.random.random(N)

    # Create transform
    transform = Affine(affine_matrix=A)

    # Check affine was passed correctly
    np.testing.assert_almost_equal(transform.affine_matrix, A)

    # Create input vector
    x = np.ones(N + 1)
    x[:-1] = np.random.random(N)

    # Apply transform and direct matrix multiplication
    result_transform = transform(x[:-1])
    result_mat_multiply = (A @ x)[:-1]

    np.testing.assert_almost_equal(result_transform, result_mat_multiply)


@pytest.mark.parametrize('dimensionality', [2, 3])
def test_affine_matrix_compose(dimensionality):
    np.random.seed(0)
    N = dimensionality
    A = np.eye(N + 1)
    A[:-1, :-1] = np.random.random((N, N))
    A[:-1, -1] = np.random.random(N)

    B = np.eye(N + 1)
    B[:-1, :-1] = np.random.random((N, N))
    B[:-1, -1] = np.random.random(N)

    # Create transform
    transform_A = Affine(affine_matrix=A)
    transform_B = Affine(affine_matrix=B)

    # Check affine was passed correctly
    np.testing.assert_almost_equal(transform_A.affine_matrix, A)
    np.testing.assert_almost_equal(transform_B.affine_matrix, B)

    # Compose transform and directly matrix multiply
    transform_C = transform_B.compose(transform_A)
    C = B @ A
    np.testing.assert_almost_equal(transform_C.affine_matrix, C)


@pytest.mark.parametrize('dimensionality', [2, 3])
def test_numpy_array_protocol(dimensionality):
    N = dimensionality
    A = np.eye(N + 1)
    A[:-1] = np.random.random((N, N + 1))
    transform = Affine(affine_matrix=A)
    np.testing.assert_almost_equal(transform.affine_matrix, A)
    np.testing.assert_almost_equal(np.asarray(transform), A)

    coords = np.random.random((20, N + 1)) * 20
    coords[:, -1] = 1
    np.testing.assert_almost_equal(
        (transform @ coords.T).T[:, :-1], transform(coords[:, :-1])
    )


@pytest.mark.parametrize('dimensionality', [2, 3])
def test_affine_matrix_inverse(dimensionality):
    np.random.seed(0)
    N = dimensionality
    A = np.eye(N + 1)
    A[:-1, :-1] = np.random.random((N, N))
    A[:-1, -1] = np.random.random(N)

    # Create transform
    transform = Affine(affine_matrix=A)

    # Check affine was passed correctly
    np.testing.assert_almost_equal(transform.affine_matrix, A)

    # Check inverse is create correctly
    np.testing.assert_almost_equal(
        transform.inverse.affine_matrix, np.linalg.inv(A)
    )


def test_repeat_shear_setting():
    """Test repeatedly setting shear with a lower triangular matrix."""
    # Note this test is needed to check lower triangular
    # decomposition of shear is working
    mat = np.eye(3)
    mat[2, 0] = 0.5
    transform = Affine(shear=mat.copy())
    # Check shear decomposed into lower triangular
    np.testing.assert_almost_equal(mat, transform.shear)

    # Set shear to same value
    transform.shear = mat.copy()
    # Check shear still decomposed into lower triangular
    np.testing.assert_almost_equal(mat, transform.shear)

    # Set shear to same value
    transform.shear = mat.copy()
    # Check shear still decomposed into lower triangular
    np.testing.assert_almost_equal(mat, transform.shear)


@pytest.mark.parametrize('dimensionality', [2, 3])
def test_composite_affine_equiv_to_affine(dimensionality):
    np.random.seed(0)
    translate = np.random.randn(dimensionality)
    scale = np.random.randn(dimensionality)
    rotate = special_ortho_group.rvs(dimensionality)
    shear = np.random.randn((dimensionality * (dimensionality - 1)) // 2)

    composite = CompositeAffine(
        translate=translate, scale=scale, rotate=rotate, shear=shear
    )
    affine = Affine(
        translate=translate, scale=scale, rotate=rotate, shear=shear
    )

    np.testing.assert_almost_equal(
        composite.affine_matrix, affine.affine_matrix
    )


def test_replace_slice_independence():
    affine = Affine(ndim=6)

    a = Affine(translate=(3, 8), rotate=33, scale=(0.75, 1.2), shear=[-0.5])
    b = Affine(translate=(2, 5), rotate=-10, scale=(1.0, 2.3), shear=[-0.0])
    c = Affine(translate=(0, 0), rotate=45, scale=(3.33, 0.9), shear=[1.5])

    affine = affine.replace_slice([1, 2], a)
    affine = affine.replace_slice([3, 4], b)
    affine = affine.replace_slice([0, 5], c)

    np.testing.assert_almost_equal(
        a.affine_matrix, affine.set_slice([1, 2]).affine_matrix
    )
    np.testing.assert_almost_equal(
        b.affine_matrix, affine.set_slice([3, 4]).affine_matrix
    )
    np.testing.assert_almost_equal(
        c.affine_matrix, affine.set_slice([0, 5]).affine_matrix
    )


def test_replace_slice_num_dimensions():
    with pytest.raises(
        ValueError, match='provided axes list and transform differ'
    ):
        Affine().replace_slice([0], Affine())


def test_affine_rotate_3d():
    a = Affine(rotate=90, ndim=3)
    npt.assert_array_almost_equal(
        np.array(
            [
                [1, 0, 0],
                [0, 0, -1],
                [0, 1, 0],
            ]
        ),
        a.rotate,
    )


@pytest.mark.parametrize('AffineType', affine_type)
def test_empty_units(AffineType):
    assert AffineType(ndim=2).units == (PIXEL, PIXEL)
    assert AffineType(ndim=3).units == (PIXEL, PIXEL, PIXEL)
    assert AffineType(ndim=2).physical_scale == (1 * PIXEL, 1 * PIXEL)
    assert AffineType(ndim=3).physical_scale == (
        1 * PIXEL,
        1 * PIXEL,
        1 * PIXEL,
    )


@pytest.mark.parametrize('AffineType', affine_type)
def test_set_units_constructor(AffineType):
    assert AffineType(ndim=2, units=('mm', 'mm')).units == (REG.mm, REG.mm)
    assert AffineType(ndim=2, units=(REG.m, REG.m)).units == (REG.m, REG.m)

    # TODO I think that we should normalize all units of same dimensionality
    # to the same registry, but this is not currently the case.
    assert AffineType(ndim=2, units=('cm', 'mm')).units == (REG.cm, REG.mm)


@pytest.mark.parametrize('AffineType', affine_type)
def test_set_units_constructor_error(AffineType):
    with pytest.raises(ValueError, match='must have length ndim'):
        AffineType(ndim=2, units=('mm', 'mm', 'mm'))

    with pytest.raises(ValueError, match='Could not find unit'):
        AffineType(ndim=2, units=('ugh', 'ugh'))


@pytest.mark.parametrize('AffineType', affine_type)
def test_set_units_error(AffineType):
    affine = AffineType(ndim=2)
    with pytest.raises(ValueError, match='must have length ndim'):
        affine.units = ('m', 'm', 'm')

    with pytest.raises(ValueError, match='Could not find unit'):
        affine.units = ('ugh', 'ugh')


@pytest.mark.parametrize('AffineType', affine_type)
def test_set_units(AffineType):
    affine = AffineType(ndim=2)
    affine.units = ('mm', 'mm')
    assert affine.units == (REG.mm, REG.mm)

    affine.units = (REG.m, REG.m)
    assert affine.units == (REG.m, REG.m)


@pytest.mark.parametrize('AffineType', affine_type)
def test_empty_axis_labels(AffineType):
    assert AffineType(ndim=2).axis_labels == ('axis -2', 'axis -1')
    assert AffineType(ndim=3).axis_labels == ('axis -3', 'axis -2', 'axis -1')


@pytest.mark.parametrize('AffineType', affine_type)
def test_set_axis_labels(AffineType):
    affine = AffineType(ndim=2)
    affine.axis_labels = ('x', 'y')
    assert affine.axis_labels == ('x', 'y')


@pytest.mark.parametrize('AffineType', affine_type)
def test_set_axis_labels_error(AffineType):
    affine = AffineType(ndim=2)
    with pytest.raises(ValueError, match='must have length ndim'):
        affine.axis_labels = ('x', 'y', 'z')


@pytest.mark.parametrize('AffineType', affine_type)
def test_set_axis_error(AffineType):
    affine = AffineType(ndim=2)
    with pytest.raises(ValueError, match='must have length ndim'):
        affine.axis_labels = ('x', 'y', 'z')



from collections.abc import Iterable
from typing import TypeVar

from napari.utils.events.containers import SelectableEventedList

T = TypeVar('T')


def _make_selectable_list_and_select_first(
    items: Iterable[T],
) -> SelectableEventedList[T]:
    selectable_list = SelectableEventedList(items)
    first = selectable_list[0]
    selectable_list.selection = [first]
    assert first in selectable_list.selection
    return selectable_list


def test_remove_discards_from_selection():
    selectable_list = _make_selectable_list_and_select_first(['a', 'b', 'c'])
    selectable_list.remove('a')
    assert 'a' not in selectable_list.selection


def test_pop_discards_from_selection():
    selectable_list = _make_selectable_list_and_select_first(['a', 'b', 'c'])
    selectable_list.pop(0)
    assert 'a' not in selectable_list.selection


def test_del_discards_from_selection():
    selectable_list = _make_selectable_list_and_select_first(['a', 'b', 'c'])
    del selectable_list[0]
    assert 'a' not in selectable_list.selection


def test_select_next():
    selectable_list = _make_selectable_list_and_select_first(['a', 'b', 'c'])
    assert 'a' in selectable_list.selection
    selectable_list.select_next()
    assert 'a' not in selectable_list.selection
    assert 'b' in selectable_list.selection


def test_select_previous():
    selectable_list = _make_selectable_list_and_select_first(['a', 'b', 'c'])
    selectable_list.selection.active = 'c'
    assert 'a' not in selectable_list.selection
    assert 'c' in selectable_list.selection
    selectable_list.select_previous()
    assert 'c' not in selectable_list.selection
    assert 'b' in selectable_list.selection


def test_shift_select_next_previous():
    selectable_list = _make_selectable_list_and_select_first(['a', 'b', 'c'])
    assert 'a' in selectable_list.selection
    selectable_list.select_next(shift=True)
    assert 'a' in selectable_list.selection
    assert 'b' in selectable_list.selection
    selectable_list.select_previous(shift=True)
    assert 'a' in selectable_list.selection
    assert 'b' not in selectable_list.selection


def test_shift_select_previous_next():
    selectable_list = _make_selectable_list_and_select_first(['a', 'b', 'c'])
    selectable_list.selection.active = 'c'
    assert 'a' not in selectable_list.selection
    assert 'c' in selectable_list.selection
    selectable_list.select_previous(shift=True)
    assert 'b' in selectable_list.selection
    assert 'c' in selectable_list.selection
    selectable_list.select_next(shift=True)
    assert 'b' not in selectable_list.selection
    assert 'c' in selectable_list.selection



from __future__ import annotations

from typing import Callable

from app_model.types import KeyCode, KeyMod

from napari.layers.points._points_constants import Mode
from napari.layers.points.points import Points
from napari.layers.utils.layer_utils import (
    register_layer_action,
    register_layer_attr_action,
)
from napari.utils.notifications import show_info
from napari.utils.translations import trans


def register_points_action(
    description: str, repeatable: bool = False
) -> Callable[[Callable], Callable]:
    return register_layer_action(Points, description, repeatable)


def register_points_mode_action(
    description: str,
) -> Callable[[Callable], Callable]:
    return register_layer_attr_action(Points, description, 'mode')


@register_points_mode_action(trans._('Transform'))
def activate_points_transform_mode(layer: Points) -> None:
    layer.mode = Mode.TRANSFORM


@register_points_mode_action(trans._('Pan/zoom'))
def activate_points_pan_zoom_mode(layer: Points) -> None:
    layer.mode = Mode.PAN_ZOOM


@register_points_mode_action(trans._('Add points'))
def activate_points_add_mode(layer: Points) -> None:
    layer.mode = Mode.ADD


@register_points_mode_action(trans._('Select points'))
def activate_points_select_mode(layer: Points) -> None:
    layer.mode = Mode.SELECT


points_fun_to_mode = [
    (activate_points_pan_zoom_mode, Mode.PAN_ZOOM),
    (activate_points_transform_mode, Mode.TRANSFORM),
    (activate_points_add_mode, Mode.ADD),
    (activate_points_select_mode, Mode.SELECT),
]


@Points.bind_key(KeyMod.CtrlCmd | KeyCode.KeyC, overwrite=True)
def copy(layer: Points) -> None:
    """Copy any selected points."""
    layer._copy_data()


@Points.bind_key(KeyMod.CtrlCmd | KeyCode.KeyV, overwrite=True)
def paste(layer: Points) -> None:
    """Paste any copied points."""
    layer._paste_data()


@register_points_action(
    trans._('Select/Deselect all points in the current view slice.'),
)
def select_all_in_slice(layer: Points) -> None:
    new_selected = set(layer._indices_view[: len(layer._view_data)])

    # If all visible points are already selected, deselect the visible points
    if new_selected & layer.selected_data == new_selected:
        layer.selected_data = layer.selected_data - new_selected
        show_info(
            trans._(
                'Deselected all points in this slice, use Shift-A to deselect all points on the layer. ({n_total} selected)',
                n_total=len(layer.selected_data),
                deferred=True,
            )
        )

    # If not all visible points are already selected, additionally select the visible points
    else:
        layer.selected_data = layer.selected_data | new_selected
        show_info(
            trans._(
                'Selected {n_new} points in this slice, use Shift-A to select all points on the layer. ({n_total} selected)',
                n_new=len(new_selected),
                n_total=len(layer.selected_data),
                deferred=True,
            )
        )


@register_points_action(
    trans._('Select/Deselect all points in the layer.'),
)
def select_all_data(layer: Points) -> None:
    # If all points are already selected, deselect all points
    if len(layer.selected_data) == len(layer.data):
        layer.selected_data = set()
        show_info(trans._('Cleared all selections.', deferred=True))

    # Select all points
    else:
        new_selected = set(range(layer.data.shape[0]))
        # Needed for the notification
        view_selected = set(layer._indices_view[: len(layer._view_data)])
        layer.selected_data = new_selected
        show_info(
            trans._(
                'Selected {n_new} points across all slices, including {n_invis} points not currently visible. ({n_total})',
                n_new=len(new_selected),
                n_invis=len(new_selected - view_selected),
                n_total=len(layer.selected_data),
                deferred=True,
            )
        )


@register_points_action(trans._('Delete selected points'))
def delete_selected_points(layer: Points) -> None:
    """Delete all selected points."""
    layer.remove_selected()



"""compatibility between newer and older python versions"""

import sys

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    # in 3.11+, using the below class in an f-string would put the enum name instead of its value
    from enum import Enum

    class StrEnum(str, Enum):
        pass



from typing import TYPE_CHECKING, Optional

import numpy as np

from napari.utils._dtype import normalize_dtype
from napari.utils.colormaps import ensure_colormap
from napari.utils.events import Event
from napari.utils.status_messages import format_float
from napari.utils.validators import _validate_increasing, validate_n_seq

validate_2_tuple = validate_n_seq(2)

if TYPE_CHECKING:
    from napari.layers._scalar_field.scalar_field import ScalarFieldBase


class IntensityVisualizationMixin:
    """A mixin that adds gamma, colormap, and contrast limits logic to Layers.

    When used, this should come before the Layer in the inheritance, e.g.:

        class Image(IntensityVisualizationMixin, Layer):
            def __init__(self):
                ...

    Note: `contrast_limits_range` is range extent available on the widget,
    and `contrast_limits` is the visible range (the set values on the widget)
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.events.add(
            contrast_limits=Event,
            contrast_limits_range=Event,
            gamma=Event,
            colormap=Event,
        )
        self._gamma = 1
        self._colormap_name = ''
        self._contrast_limits_msg = ''
        self._contrast_limits: tuple[Optional[float], Optional[float]] = (
            None,
            None,
        )
        self._contrast_limits_range: tuple[
            Optional[float], Optional[float]
        ] = (None, None)
        self._auto_contrast_source = 'slice'
        self._keep_auto_contrast = False

    def reset_contrast_limits(self: 'ScalarFieldBase', mode=None):
        """Scale contrast limits to data range"""
        mode = mode or self._auto_contrast_source
        self.contrast_limits = self._calc_data_range(mode)

    def _calc_data_range(self, mode):
        raise NotImplementedError

    def reset_contrast_limits_range(self, mode=None):
        """Scale contrast limits range to data type if dtype is an integer,
        or use the current maximum data range otherwise.
        """
        dtype = normalize_dtype(self.dtype)
        if np.issubdtype(dtype, np.integer):
            info = np.iinfo(dtype)
            self.contrast_limits_range = (info.min, info.max)
        else:
            mode = mode or self._auto_contrast_source
            self.contrast_limits_range = self._calc_data_range(mode)

    @property
    def colormap(self):
        """napari.utils.Colormap: colormap for luminance images."""
        return self._colormap

    @colormap.setter
    def colormap(self, colormap):
        self._set_colormap(colormap)

    def _set_colormap(self, colormap):
        self._colormap = ensure_colormap(colormap)
        self._update_thumbnail()
        self.events.colormap()

    @property
    def colormaps(self):
        """tuple of str: names of available colormaps."""
        return tuple(self._colormaps.keys())

    @property
    def contrast_limits(self):
        """list of float: Limits to use for the colormap."""
        return list(self._contrast_limits)

    @contrast_limits.setter
    def contrast_limits(self, contrast_limits):
        validate_2_tuple(contrast_limits)
        _validate_increasing(contrast_limits)
        self._contrast_limits_msg = (
            format_float(contrast_limits[0])
            + ', '
            + format_float(contrast_limits[1])
        )
        self._contrast_limits = contrast_limits
        # make sure range slider is big enough to fit range
        newrange = list(self.contrast_limits_range)
        newrange[0] = min(newrange[0], contrast_limits[0])
        newrange[1] = max(newrange[1], contrast_limits[1])
        self.contrast_limits_range = newrange
        self._update_thumbnail()
        self.events.contrast_limits()

    @property
    def contrast_limits_range(self):
        """The current valid range of the contrast limits."""
        return list(self._contrast_limits_range)

    @contrast_limits_range.setter
    def contrast_limits_range(self, value):
        """Set the valid range of the contrast limits.
        If either value is "None", the current range will be preserved.
        If the range overlaps the current contrast limits, the range will be set
        requested and there will be no change the contrast limits.
        If the requested contrast range limits are completely outside the
        current contrast limits, the range will be set as requested and the
        contrast limits will be reset to the new range.
        """
        validate_2_tuple(value)
        _validate_increasing(value)
        if list(value) == self.contrast_limits_range:
            return

        # if either value is "None", it just preserves the current range
        current_range = self.contrast_limits_range
        value = list(value)  # make sure it is mutable
        for i in range(2):
            value[i] = current_range[i] if value[i] is None else value[i]
        self._contrast_limits_range = value
        self.events.contrast_limits_range()

        # make sure that the contrast limits fit within the new range
        # this also serves the purpose of emitting events.contrast_limits()
        # and updating the views/controllers
        if hasattr(self, '_contrast_limits') and any(self._contrast_limits):
            clipped_limits = np.clip(self.contrast_limits, *value)
            if clipped_limits[0] < clipped_limits[1]:
                self.contrast_limits = tuple(clipped_limits)
            else:
                self.contrast_limits = tuple(value)

    @property
    def gamma(self):
        return self._gamma

    @gamma.setter
    def gamma(self, value):
        self._gamma = float(value)
        self._update_thumbnail()
        self.events.gamma()



import sys
import threading
import warnings

import pytest

from napari._tests.utils import DEFAULT_TIMEOUT_SECS
from napari.utils.notifications import (
    Notification,
    notification_manager,
    show_debug,
    show_error,
    show_info,
    show_warning,
)


# capsys fixture comes from pytest
# https://docs.pytest.org/en/stable/logging.html#caplog-fixture
def test_keyboard_interupt_handler(capsys):
    with pytest.raises(SystemExit):
        notification_manager.receive_error(
            KeyboardInterrupt, KeyboardInterrupt(), None
        )


class PurposefulException(Exception):
    pass


def test_notification_repr_has_message():
    assert "='this is the message'" in repr(
        Notification('this is the message')
    )


def test_notification_manager_no_gui(monkeypatch):
    """
    Direct test of the notification manager.

    This does not test the integration with the gui, but test that the
    notification manager itself can receive a info, warning or error.
    """
    previous_exhook = sys.excepthook
    with notification_manager:
        notification_manager.records.clear()
        # save all of the events that get emitted
        store: list[Notification] = []
        notification_manager.notification_ready.connect(store.append)

        show_info('this is one way of showing an information message')
        assert (
            len(notification_manager.records) == 1
        ), notification_manager.records
        assert store[-1].type == 'info'

        notification_manager.receive_info(
            'This is another information message'
        )
        assert len(notification_manager.records) == 2
        assert len(store) == 2
        assert store[-1].type == 'info'

        # test that exceptions that go through sys.excepthook are catalogued

        with pytest.raises(PurposefulException):
            raise PurposefulException('this is an exception')

        # pytest intercepts the error, so we can manually call sys.excepthook
        assert sys.excepthook == notification_manager.receive_error
        try:
            raise ValueError('a')
        except ValueError:
            sys.excepthook(*sys.exc_info())
        assert len(notification_manager.records) == 3
        assert len(store) == 3
        assert store[-1].type == 'error'

        # test that warnings that go through showwarning are catalogued
        # again, pytest intercepts this, so just manually trigger:
        assert warnings.showwarning == notification_manager.receive_warning
        warnings.showwarning(
            UserWarning('this is a warning'), UserWarning, __file__, 83
        )
        assert len(notification_manager.records) == 4
        assert store[-1].type == 'warning'

        show_error('This is an error')
        assert len(notification_manager.records) == 5
        assert store[-1].type == 'error'

        show_warning('This is a warning')
        assert len(notification_manager.records) == 6
        assert store[-1].type == 'warning'

        show_debug('This is a debug')
        assert len(notification_manager.records) == 7
        assert store[-1].type == 'debug'

    # make sure we've restored the except hook
    assert sys.excepthook == previous_exhook

    assert all(isinstance(x, Notification) for x in store)


def test_notification_manager_no_gui_with_threading():
    """
    Direct test of the notification manager.

    This does not test the integration with the gui, but test that
    exceptions and warnings from threads are correctly captured.
    """

    def _warn():
        warnings.showwarning(
            UserWarning('this is a warning'), UserWarning, __file__, 116
        )

    def _raise():
        with pytest.raises(PurposefulException):
            raise PurposefulException('this is an exception')

    previous_threading_exhook = threading.excepthook

    with notification_manager:
        notification_manager.records.clear()
        # save all of the events that get emitted
        store: list[Notification] = []
        notification_manager.notification_ready.connect(store.append)

        # Test exception inside threads
        assert (
            threading.excepthook == notification_manager.receive_thread_error
        )

        exception_thread = threading.Thread(target=_raise)
        exception_thread.start()
        exception_thread.join(timeout=DEFAULT_TIMEOUT_SECS)

        try:
            raise ValueError('a')
        except ValueError:
            threading.excepthook(sys.exc_info())

        assert len(notification_manager.records) == 1
        assert store[-1].type == 'error'

        # Test warning inside threads
        assert warnings.showwarning == notification_manager.receive_warning
        warning_thread = threading.Thread(target=_warn)
        warning_thread.start()
        warning_thread.join(timeout=DEFAULT_TIMEOUT_SECS)

        assert len(notification_manager.records) == 2
        assert store[-1].type == 'warning'

    # make sure we've restored the threading except hook
    assert threading.excepthook == previous_threading_exhook

    assert all(isinstance(x, Notification) for x in store)


def test_notification_manager_no_warning_duplication():
    def fun():
        warnings.showwarning(
            UserWarning('This is a warning'),
            category=UserWarning,
            filename=__file__,
            lineno=166,
        )

    with notification_manager:
        notification_manager.records.clear()
        # save all of the events that get emitted
        store: list[Notification] = []
        notification_manager.notification_ready.connect(store.append)

        fun()
        assert len(notification_manager.records) == 1
        assert store[-1].type == 'warning'

        fun()
        assert len(notification_manager.records) == 1



import pytest

from napari._qt.containers import (
    QtListModel,
    QtListView,
    QtNodeTreeModel,
    QtNodeTreeView,
    create_view,
)
from napari.utils.events.containers import SelectableEventedList
from napari.utils.tree import Group, Node


class T(Node):
    def __init__(self, x) -> None:
        self.x = x


@pytest.mark.parametrize(
    ('cls', 'exView', 'exModel'),
    [
        (SelectableEventedList, QtListView, QtListModel),
        (Group, QtNodeTreeView, QtNodeTreeModel),
    ],
)
def test_factory(qtbot, cls, exView, exModel):
    a = cls([T(1), T(2)])
    view = create_view(a)
    qtbot.addWidget(view)
    assert isinstance(view, exView)
    assert isinstance(view.model(), exModel)



from vispy.scene.visuals import Volume as BaseVolume

from napari._vispy.visuals.util import TextureMixin
from napari.layers.labels._labels_constants import IsoCategoricalGradientMode

FUNCTION_DEFINITIONS = """
uniform bool u_iso_gradient;

// the tolerance for testing equality of floats with floatEqual and floatNotEqual
const float equality_tolerance = 1e-8;

bool floatNotEqual(float val1, float val2)
{
    // check if val1 and val2 are not equal
    bool not_equal = abs(val1 - val2) > equality_tolerance;

    return not_equal;
}

bool floatEqual(float val1, float val2)
{
    // check if val1 and val2 are equal
    bool equal = abs(val1 - val2) < equality_tolerance;

    return equal;
}


// the background value for the iso_categorical shader
const float categorical_bg_value = 0;

int detectAdjacentBackground(float val_neg, float val_pos)
{
    // determine if the adjacent voxels along an axis are both background
    int adjacent_bg = int( floatEqual(val_neg, categorical_bg_value) );
    adjacent_bg = adjacent_bg * int( floatEqual(val_pos, categorical_bg_value) );
    return adjacent_bg;
}

vec3 calculateGradient(vec3 loc, vec3 step, out int n_bg_borders) {
    // calculate gradient within the volume by finite differences

    n_bg_borders = 0;
    vec3 G = vec3(0.0);

    float prev;
    float next;

    prev = colorToVal($get_data(loc - vec3(step.x, 0, 0)));
    next = colorToVal($get_data(loc + vec3(step.x, 0, 0)));
    n_bg_borders += detectAdjacentBackground(prev, next);
    G.x = next - prev;

    prev = colorToVal($get_data(loc - vec3(0, step.y, 0)));
    next = colorToVal($get_data(loc + vec3(0, step.y, 0)));
    n_bg_borders += detectAdjacentBackground(prev, next);
    G.y = next - prev;

    prev = colorToVal($get_data(loc - vec3(0, 0, step.z)));
    next = colorToVal($get_data(loc + vec3(0, 0, step.z)));
    n_bg_borders += detectAdjacentBackground(prev, next);
    G.z = next - prev;

    return G;
}

vec3 calculateIsotropicGradient(vec3 loc, vec3 step) {
    // calculate gradient within the volume by finite differences
    // using a 3D sobel-feldman convolution kernel

    // the kernel here is a 3x3 cube, centered on the sample at `loc`
    // the kernel for G.z looks like this:

    // [ +1 +2 +1 ]
    // [ +2 +4 +2 ]    <-- "loc - step.z" is in the center
    // [ +1 +2 +1 ]

    // [  0  0  0 ]
    // [  0  0  0 ]    <-- "loc" is in the center
    // [  0  0  0 ]

    // [ -1 -2 -1 ]
    // [ -2 -4 -2 ]    <-- "loc + step.z" is in the center
    // [ -1 -2 -1 ]

    // kernels for G.x and G.y similar, but transposed
    // see https://en.wikipedia.org/wiki/Sobel_operator#Extension_to_other_dimensions

    vec3 G = vec3(0.0);

    for (int i=-1; i <= 1; i++) {
        for (int j=-1; j <= 1; j++) {
            for (int k=-1; k <= 1; k++) {
                float val = colorToVal($get_data(loc + vec3(i, j, k) * step));
                G.x += val * -float(i) *
                    (1 + float(j == 0 || k == 0) + 2 * float(j == 0 && k == 0));
                G.y += val * -float(j) *
                    (1 + float(i == 0 || k == 0) + 2 * float(i == 0 && k == 0));
                G.z += val * -float(k) *
                    (1 + float(i == 0 || j == 0) + 2 * float(i == 0 && j == 0));
            }
        }
    }

    return G;
}

vec4 calculateShadedCategoricalColor(vec4 betterColor, vec3 loc, vec3 step)
{
    // Calculate color by incorporating ambient and diffuse lighting
    vec4 color0 = $get_data(loc);
    vec4 color1;
    vec4 color2;
    float val0 = colorToVal(color0);
    float val1 = 0;
    float val2 = 0;
    int n_bg_borders = 0;

    // View direction
    vec3 V = normalize(view_ray);

    // Calculate normal vector from gradient
    vec3 N;
    if (u_iso_gradient) {
        N = calculateIsotropicGradient(loc, step);
    } else {
        N = calculateGradient(loc, step, n_bg_borders);
    }

    // Normalize and flip normal so it points towards viewer
    N = normalize(N);
    float Nselect = float(dot(N,V) > 0.0);
    N = (2.0*Nselect - 1.0) * N;  // ==  Nselect * N - (1.0-Nselect)*N;

    // Init colors
    vec4 ambient_color = vec4(0.0, 0.0, 0.0, 0.0);
    vec4 diffuse_color = vec4(0.0, 0.0, 0.0, 0.0);
    vec4 final_color;

    // todo: allow multiple light, define lights on viewvox or subscene
    int nlights = 1;
    for (int i=0; i<nlights; i++)
    {
        // Get light direction (make sure to prevent zero devision)
        vec3 L = normalize(view_ray);  //lightDirs[i];
        float lightEnabled = float( length(L) > 0.0 );
        L = normalize(L+(1.0-lightEnabled));

        // Calculate lighting properties
        float lambertTerm = clamp( dot(N,L), 0.0, 1.0 );
        if (n_bg_borders > 0) {
            // to fix dim pixels due to poor normal estimation,
            // we give a default lambda to pixels surrounded by background
            lambertTerm = 0.5;
        }

        // Calculate mask
        float mask1 = lightEnabled;

        // Calculate colors
        ambient_color +=  mask1 * u_ambient;  // * gl_LightSource[i].ambient;
        diffuse_color +=  mask1 * lambertTerm;
    }

    // Calculate final color by componing different components
    final_color = betterColor * ( ambient_color + diffuse_color);
    final_color.a = betterColor.a;

    // Done
    return final_color;
}
"""

ISO_CATEGORICAL_SNIPPETS = {
    'before_loop': """
        vec4 color3 = vec4(0.0);  // final color
        vec3 dstep = 1.5 / u_shape;  // step to sample derivative, set to match iso shader
        gl_FragColor = vec4(0.0);
        bool discard_fragment = true;
        vec4 label_id = vec4(0.0);
        """,
    'in_loop': """
        // check if value is different from the background value
        if ( floatNotEqual(val, categorical_bg_value) ) {
            // Take the last interval in smaller steps
            vec3 iloc = loc - step;
            for (int i=0; i<10; i++) {
                label_id = $get_data(iloc);
                color = sample_label_color(label_id.r);
                if (floatNotEqual(color.a, 0) ) {
                    // fully transparent color is considered as background, see napari/napari#5227
                    // when the value mapped to non-transparent color is reached
                    // calculate the shaded color (apply lighting effects)
                    color = calculateShadedCategoricalColor(color, iloc, dstep);
                    gl_FragColor = color;

                    // set the variables for the depth buffer
                    frag_depth_point = iloc * u_shape;
                    discard_fragment = false;

                    iter = nsteps;
                    break;
                }
                iloc += step * 0.1;
            }
        }
        """,
    'after_loop': """
        if (discard_fragment)
            discard;
        """,
}

TRANSLUCENT_CATEGORICAL_SNIPPETS = {
    'before_loop': """
        vec4 color3 = vec4(0.0);  // final color
        gl_FragColor = vec4(0.0);
        bool discard_fragment = true;
        vec4 label_id = vec4(0.0);
        """,
    'in_loop': """
        // check if value is different from the background value
        if ( floatNotEqual(val, categorical_bg_value) ) {
            // Take the last interval in smaller steps
            vec3 iloc = loc - step;
            for (int i=0; i<10; i++) {
                label_id = $get_data(iloc);
                color = sample_label_color(label_id.r);
                if (floatNotEqual(color.a, 0) ) {
                    // fully transparent color is considered as background, see napari/napari#5227
                    // when the value mapped to non-transparent color is reached
                    // calculate the color (apply lighting effects)
                    gl_FragColor = color;

                    // set the variables for the depth buffer
                    frag_depth_point = iloc * u_shape;
                    discard_fragment = false;

                    iter = nsteps;
                    break;
                }
                iloc += step * 0.1;
            }
        }
        """,
    'after_loop': """
        if (discard_fragment)
            discard;
        """,
}

shaders = BaseVolume._shaders.copy()
before, after = shaders['fragment'].split('void main()')
shaders['fragment'] = before + FUNCTION_DEFINITIONS + 'void main()' + after

rendering_methods = BaseVolume._rendering_methods.copy()
rendering_methods['iso_categorical'] = ISO_CATEGORICAL_SNIPPETS
rendering_methods['translucent_categorical'] = TRANSLUCENT_CATEGORICAL_SNIPPETS


class Volume(TextureMixin, BaseVolume):
    """This class extends the vispy Volume visual to add categorical isosurface rendering."""

    # add the new rendering method to the snippets dict
    _shaders = shaders
    _rendering_methods = rendering_methods

    def __init__(self, *args, **kwargs) -> None:  # type: ignore [no-untyped-def]
        super().__init__(*args, **kwargs)
        self.unfreeze()
        self.iso_gradient_mode = IsoCategoricalGradientMode.FAST.value
        self.freeze()

    @property
    def iso_gradient_mode(self) -> str:
        return str(self._iso_gradient_mode)

    @iso_gradient_mode.setter
    def iso_gradient_mode(self, value: str) -> None:
        self._iso_gradient_mode = IsoCategoricalGradientMode(value)
        self.shared_program['u_iso_gradient'] = (
            self._iso_gradient_mode == IsoCategoricalGradientMode.SMOOTH
        )
        self.update()



import numpy as np
import pytest

from napari._tests.utils import (
    add_layer_by_type,
    layer_test_data,
)
from napari.components._viewer_key_bindings import (
    hold_for_pan_zoom,
    rotate_layers,
    show_only_layer_above,
    show_only_layer_below,
    toggle_selected_visibility,
    toggle_theme,
    toggle_unselected_visibility,
)
from napari.components.viewer_model import ViewerModel
from napari.layers.points import Points
from napari.settings import get_settings
from napari.utils.theme import available_themes, get_system_theme


def test_theme_toggle_keybinding():
    viewer = ViewerModel()
    assert viewer.theme == get_settings().appearance.theme
    assert viewer.theme != 'light'
    toggle_theme(viewer)
    # toggle_theme should not change settings
    assert get_settings().appearance.theme != 'light'
    # toggle_theme should change the viewer theme
    assert viewer.theme == 'light'
    # ensure toggle_theme loops through all themes
    initial_theme = viewer.theme
    number_of_actual_themes = len(available_themes())
    if 'system' in available_themes():
        number_of_actual_themes = len(available_themes()) - 1
    for _i in range(number_of_actual_themes):
        current_theme = viewer.theme
        toggle_theme(viewer)
        # theme should have changed
        assert viewer.theme != current_theme
        # toggle_theme should toggle only actual themes
        assert viewer.theme != 'system'
    # ensure we're back at the initial theme
    assert viewer.theme == initial_theme


def test_theme_toggle_from_system_theme():
    get_settings().appearance.theme = 'system'
    viewer = ViewerModel()
    assert viewer.theme == 'system'
    actual_initial_theme = get_system_theme()
    toggle_theme(viewer)
    # ensure that theme has changed
    assert viewer.theme != actual_initial_theme
    assert viewer.theme != 'system'
    number_of_actual_themes = len(available_themes())
    if 'system' in available_themes():
        number_of_actual_themes = len(available_themes()) - 1
    for _i in range(number_of_actual_themes - 1):  # we've already toggled once
        current_theme = viewer.theme
        toggle_theme(viewer)
        # theme should have changed
        assert viewer.theme != current_theme
        # toggle_theme should toggle only actual themes
        assert viewer.theme != 'system'
    # ensure we have looped back to whatever system was
    assert viewer.theme == actual_initial_theme


def test_hold_for_pan_zoom():
    viewer = ViewerModel()
    data = [[1, 3], [8, 4], [10, 10], [15, 4]]
    layer = Points(data, size=1)
    viewer.layers.append(layer)
    layer.mode = 'transform'

    viewer.layers.selection.active = viewer.layers[0]
    gen = hold_for_pan_zoom(viewer)
    assert layer.mode == 'transform'
    next(gen)
    assert layer.mode == 'pan_zoom'
    with pytest.raises(StopIteration):
        next(gen)
    assert layer.mode == 'transform'


def test_selected_visibility_toggle():
    viewer = make_viewer_with_three_layers()
    viewer.layers.selection.active = viewer.layers[0]
    assert viewer.layers[0].visible
    assert viewer.layers[1].visible
    assert viewer.layers[2].visible
    toggle_selected_visibility(viewer)
    assert not viewer.layers[0].visible
    assert viewer.layers[1].visible
    assert viewer.layers[2].visible
    toggle_selected_visibility(viewer)
    assert viewer.layers[0].visible
    assert viewer.layers[1].visible
    assert viewer.layers[2].visible


def test_unselected_visibility_toggle():
    viewer = make_viewer_with_three_layers()
    viewer.layers.selection.active = viewer.layers[0]
    assert viewer.layers[0].visible
    assert viewer.layers[1].visible
    assert viewer.layers[2].visible
    toggle_unselected_visibility(viewer)
    assert viewer.layers[0].visible
    assert not viewer.layers[1].visible
    assert not viewer.layers[2].visible
    toggle_unselected_visibility(viewer)
    assert viewer.layers[0].visible
    assert viewer.layers[1].visible
    assert viewer.layers[2].visible


def test_show_only_layer_above():
    viewer = make_viewer_with_three_layers()
    viewer.layers.selection.active = viewer.layers[0]
    assert viewer.layers[0].visible
    assert viewer.layers[1].visible
    assert viewer.layers[2].visible
    show_only_layer_above(viewer)
    assert not viewer.layers[0].visible
    assert viewer.layers[1].visible
    assert not viewer.layers[2].visible
    show_only_layer_above(viewer)
    assert not viewer.layers[0].visible
    assert not viewer.layers[1].visible
    assert viewer.layers[2].visible


def test_show_only_layer_below():
    viewer = make_viewer_with_three_layers()
    viewer.layers.selection.active = viewer.layers[2]
    assert viewer.layers[0].visible
    assert viewer.layers[1].visible
    assert viewer.layers[2].visible
    show_only_layer_below(viewer)
    assert not viewer.layers[2].visible
    assert viewer.layers[1].visible
    assert not viewer.layers[0].visible
    show_only_layer_below(viewer)
    assert not viewer.layers[2].visible
    assert not viewer.layers[1].visible
    assert viewer.layers[0].visible


@pytest.mark.parametrize(('layer_class', 'data', 'ndim'), layer_test_data)
def test_rotate_layers(layer_class, data, ndim):
    """Test rotate layers works with all layer types/data"""
    viewer = ViewerModel()
    layer = add_layer_by_type(viewer, layer_class, data, visible=True)
    np.testing.assert_array_equal(
        layer.affine.rotate, np.eye(ndim, dtype=float)
    )
    rotate_layers(viewer)
    np.testing.assert_array_equal(
        layer.affine.rotate[-2:, -2:], np.array([[0, -1], [1, 0]], dtype=float)
    )


def test_rotate_layers_in_3D():
    """Test that rotate layers is disabled in 3D viewer mode"""
    viewer = ViewerModel()
    layer = add_layer_by_type(
        viewer, Points, np.array([[1, 2, 3]]), visible=True
    )
    initial_rotation_matrix = layer.affine.rotate[-2:, -2:]
    viewer.dims.ndisplay = 3
    assert viewer.dims.ndisplay == 3
    rotate_layers(viewer)
    # with ndisplay == 3 rotation is disabled, the matrix should not have changed
    np.testing.assert_array_equal(
        layer.affine.rotate[-2:, -2:], initial_rotation_matrix
    )


def make_viewer_with_three_layers():
    """Helper function to create a viewer with three layers"""
    viewer = ViewerModel()
    layer1 = Points()
    layer2 = Points()
    layer3 = Points()
    viewer.layers.append(layer1)
    viewer.layers.append(layer2)
    viewer.layers.append(layer3)
    return viewer



from dataclasses import dataclass, field
from typing import Any

import numpy as np
import numpy.typing as npt

from napari.layers.base._slice import _next_request_id
from napari.layers.points._points_constants import PointsProjectionMode
from napari.layers.utils._slice_input import _SliceInput, _ThickNDSlice


@dataclass(frozen=True)
class _PointSliceResponse:
    """Contains all the output data of slicing an points layer.

    Attributes
    ----------
    indices : array like
        Indices of the sliced Points data.
    scale: array like or none
        Used to scale the sliced points for visualization.
        Should be broadcastable to indices.
    slice_input : _SliceInput
        Describes the slicing plane or bounding box in the layer's dimensions.
    request_id : int
        The identifier of the request from which this was generated.
    """

    indices: np.ndarray = field(repr=False)
    scale: Any = field(repr=False)
    slice_input: _SliceInput
    request_id: int


@dataclass(frozen=True)
class _PointSliceRequest:
    """A callable that stores all the input data needed to slice a Points layer.

    This should be treated a deeply immutable structure, even though some
    fields can be modified in place. It is like a function that has captured
    all its inputs already.

    In general, the calling an instance of this may take a long time, so you may
    want to run it off the main thread.

    Attributes
    ----------
    dims : _SliceInput
        Describes the slicing plane or bounding box in the layer's dimensions.
    data : Any
        The layer's data field, which is the main input to slicing.
    data_slice : _ThickNDSlice
        The slicing coordinates and margins in data space.
    size : array like
        Size of each point. This is used in calculating visibility.
    others
        See the corresponding attributes in `Layer` and `Points`.
    """

    slice_input: _SliceInput
    data: Any = field(repr=False)
    data_slice: _ThickNDSlice = field(repr=False)
    projection_mode: PointsProjectionMode
    size: Any = field(repr=False)
    out_of_slice_display: bool = field(repr=False)
    id: int = field(default_factory=_next_request_id)

    def __call__(self) -> _PointSliceResponse:
        # Return early if no data
        if len(self.data) == 0:
            return _PointSliceResponse(
                indices=np.array([], dtype=int),
                scale=np.empty(0),
                slice_input=self.slice_input,
                request_id=self.id,
            )

        not_disp = list(self.slice_input.not_displayed)
        if not not_disp:
            # If we want to display everything, then use all indices.
            # scale is only impacted by not displayed data, therefore 1
            return _PointSliceResponse(
                indices=np.arange(len(self.data), dtype=int),
                scale=1,
                slice_input=self.slice_input,
                request_id=self.id,
            )

        slice_indices, scale = self._get_slice_data(not_disp)

        return _PointSliceResponse(
            indices=slice_indices,
            scale=scale,
            slice_input=self.slice_input,
            request_id=self.id,
        )

    def _get_slice_data(self, not_disp: list[int]) -> tuple[npt.NDArray, int]:
        data = self.data[:, not_disp]
        scale = 1

        point, m_left, m_right = self.data_slice[not_disp].as_array()

        if self.projection_mode == 'none':
            low = point.copy()
            high = point.copy()
        else:
            low = point - m_left
            high = point + m_right

        # assume slice thickness of 1 in data pixels
        # (same as before thick slices were implemented)
        too_thin_slice = np.isclose(high, low)
        low[too_thin_slice] -= 0.5
        high[too_thin_slice] += 0.5

        inside_slice = np.all((data >= low) & (data <= high), axis=1)
        slice_indices = np.where(inside_slice)[0].astype(int)

        if self.out_of_slice_display and self.slice_input.ndim > 2:
            sizes = self.size[:, np.newaxis] / 2

            # add out of slice points with progressively lower sizes
            dist_from_low = np.abs(data - low)
            dist_from_high = np.abs(data - high)
            distances = np.minimum(dist_from_low, dist_from_high)
            # anything inside the slice is at distance 0
            distances[inside_slice] = 0

            # display points that "spill" into the slice
            matches = np.all(distances <= sizes, axis=1)
            if not np.any(matches):
                return np.empty(0, dtype=int), 1
            size_match = sizes[matches]
            # rescale size of spilling points based on how much they do
            scale_per_dim = (size_match - distances[matches]) / size_match
            scale = np.prod(scale_per_dim, axis=1)
            slice_indices = np.where(matches)[0].astype(int)

        return slice_indices, scale



import logging

from qtpy import API_NAME
from vispy import app

# set vispy application to the appropriate qt backend
app.use_app(API_NAME)
del app

# set vispy logger to show warning and errors only
vispy_logger = logging.getLogger('vispy')
vispy_logger.setLevel(logging.WARNING)


from napari._vispy.camera import VispyCamera
from napari._vispy.canvas import VispyCanvas
from napari._vispy.overlays.axes import VispyAxesOverlay
from napari._vispy.overlays.interaction_box import (
    VispySelectionBoxOverlay,
    VispyTransformBoxOverlay,
)
from napari._vispy.overlays.labels_polygon import VispyLabelsPolygonOverlay
from napari._vispy.overlays.scale_bar import VispyScaleBarOverlay
from napari._vispy.overlays.text import VispyTextOverlay
from napari._vispy.utils.quaternion import quaternion2euler_degrees
from napari._vispy.utils.visual import create_vispy_layer, create_vispy_overlay

__all__ = [
    'VispyCamera',
    'VispyCanvas',
    'VispyAxesOverlay',
    'VispySelectionBoxOverlay',
    'VispyScaleBarOverlay',
    'VispyTransformBoxOverlay',
    'VispyTextOverlay',
    'VispyLabelsPolygonOverlay',
    'quaternion2euler_degrees',
    'create_vispy_layer',
    'create_vispy_overlay',
]



from pathlib import Path
from typing import Any, Optional

from napari.settings._base import _NOT_SET
from napari.settings._napari_settings import (
    CURRENT_SCHEMA_VERSION,
    NapariSettings,
)
from napari.utils.translations import trans

__all__ = ['NapariSettings', 'get_settings', 'CURRENT_SCHEMA_VERSION']


class _SettingsProxy:
    """Backwards compatibility layer."""

    def __getattribute__(self, name) -> Any:
        return getattr(get_settings(), name)


# deprecated
SETTINGS = _SettingsProxy()

# private global object
# will be populated on first call of get_settings
_SETTINGS: Optional[NapariSettings] = None


def get_settings(path=_NOT_SET) -> NapariSettings:
    """
    Get settings for a given path.

    Parameters
    ----------
    path : Path, optional
        The path to read/write the settings from.

    Returns
    -------
    SettingsManager
        The settings manager.

    Notes
    -----
    The path can only be set once per session.
    """
    global _SETTINGS

    if _SETTINGS is None:
        if path is not _NOT_SET:
            path = Path(path).resolve() if path is not None else None
        _SETTINGS = NapariSettings(config_path=path)
    elif path is not _NOT_SET:
        import inspect

        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        raise RuntimeError(
            trans._(
                'The path can only be set once per session. Settings called from {calframe}',
                deferred=True,
                calframe=calframe[1][3],
            )
        )

    return _SETTINGS



"""Monitor class.

The Monitor class wraps the MonitorServer and MonitorApi. One reason
for having a wrapper class is that so the rest of napari does not
need to import any multiprocessing code unless actually using
the monitor.
"""

import errno
import json
import logging
import os
from pathlib import Path
from typing import Optional

from napari.utils.translations import trans

LOGGER = logging.getLogger('napari.monitor')

# If False monitor is disabled even if we meet all other requirements.
ENABLE_MONITOR = True


def _load_config(path: str) -> dict:
    """Load the JSON formatted config file.

    Parameters
    ----------
    path : str
        The path of the JSON file we should load.

    Returns
    -------
    dict
        The parsed data from the JSON file.
    """
    path = Path(path).expanduser()
    if not path.exists():
        raise FileNotFoundError(
            errno.ENOENT,
            trans._(
                'Monitor: Config file not found: {path}',
                deferred=True,
                path=path,
            ),
        )

    with path.open() as infile:
        return json.load(infile)


def _load_monitor_config() -> Optional[dict]:
    """Return the MonitorService config file data, or None.

    Returns
    -------
    Optional[dict]
        The parsed config file data or None if no config.
    """
    # We shouldn't even call into this file unless NAPARI_MON is defined
    # but check to be sure.
    value = os.getenv('NAPARI_MON')
    if value in [None, '0']:
        return None

    return _load_config(value)


def _setup_logging(config: dict) -> None:
    """Log "napari.monitor" messages to the configured file.

    Parameters
    ----------
    config : dict
        Monitor configuration
    """
    try:
        log_path = config['log_path']
    except KeyError:
        return  # No log file.

    # Nuke/reset log for now.
    # Path(log_path).unlink()

    fh = logging.FileHandler(log_path)
    LOGGER.addHandler(fh)
    LOGGER.setLevel(logging.DEBUG)
    LOGGER.info('Writing to log path %s', log_path)


def _get_monitor_config() -> Optional[dict]:
    """Create and return the configuration for the MonitorService.

    The routine might return None for one serveral reasons:
    1) We're not running under Python 3.9 or now.
    2) The monitor is explicitly disable, ENABLED_MONITOR is False.
    3) The NAPARI_MON environment variable is not defined.
    4) The NAPARI_MON config file cannot be found and parsed.

    Returns
    -------
    Optional[dict]
        The configuration for the MonitorService.
    """

    if not ENABLE_MONITOR:
        logging.warning('Monitor: not starting, disabled')
        return None

    # The NAPARI_MON environment variable points to our config file.
    config = _load_monitor_config()

    if config is None:
        logging.warning('Monitor: not starting, no usable config file')
        return None

    return config


class Monitor:
    """Wraps the monitor service.

    We can't start the monitor service at import time. Under the hood the
    multiprocessing complains about a "partially started process".

    Instead someone must call our start() method explicitly once the
    process has fully started.
    """

    def __init__(self) -> None:
        # Both are set when start() is called, and only if we have
        # a parseable config file, have Python 3.9, etc.
        self._service = None
        self._api = None
        self._running = False

    def __nonzero__(self) -> bool:
        """Return True if the service is running.

        So that callers can do:

            if monitor:
                monitor.add(...)
        """
        return self._running

    @property
    def run_command_event(self):
        """The MonitorAPI fires this event for commands from clients."""
        return self._api.events.run_command

    def start(self) -> bool:
        """Start the monitor service, if it hasn't been started already.

        Returns
        -------
        bool
            True if we started the service or it was already started.
        """
        if self._running:
            return True  # It was already started.

        config = _get_monitor_config()

        if config is None:
            return False  # Can't start without config.

        _setup_logging(config)

        # Late imports so no multiprocessing modules are even
        # imported unless we are going to start the service.
        from napari.components.experimental.monitor._api import MonitorApi
        from napari.components.experimental.monitor._service import (
            MonitorService,
        )

        # Create the API first. It will register our callbacks, then
        # we start the manager that will serve those callbacks.
        self._api = MonitorApi()

        # Now we can start our service.
        self._service = MonitorService(config, self._api.manager)

        self._running = True
        return True  # We started the service.

    def stop(self) -> None:
        """Stop the monitor service."""
        if not self._running:
            return

        self._api.stop()
        self._api = None

        self._service.stop()
        self._service = None

        self._running = False

    def on_poll(self, event) -> None:
        """The QtPoll object polls us.

        Probably we could get rid of polling by creating a thread that
        blocks waiting for client messages. Then it posts those messages as
        Qt Events. So the GUI doesn't block, but gracefully handles
        incoming messages as Qt events.
        """
        if self._running:
            self._api.poll()

        # Handle the event to say "keep polling us".
        event.handled = True

    def add_data(self, data) -> None:
        """Add data to the monitor service.

        Caller should use this pattern:
            if monitor:
                monitor.add(...)

        So no time wasted assembling the dict unless the monitor is running.
        """
        if self._running:
            self._api.add_napari_data(data)

    def send_message(self, message: dict) -> None:
        """Send a message to shared memory clients.

        Parameters
        ----------
        message : dict
            Post this message to clients.
        """
        if self._running:
            self._api.send_napari_message(message)


monitor = Monitor()



import numpy as np

from napari.components import Camera


def test_camera():
    """Test camera."""
    camera = Camera()
    assert camera.center == (0, 0, 0)
    assert camera.zoom == 1
    assert camera.angles == (0, 0, 90)

    center = (10, 20, 30)
    camera.center = center
    assert camera.center == center
    assert camera.angles == (0, 0, 90)

    zoom = 200
    camera.zoom = zoom
    assert camera.zoom == zoom

    angles = (20, 90, 45)
    camera.angles = angles
    assert camera.angles == angles


def test_calculate_view_direction_3d():
    """Check that view direction is calculated properly from camera angles."""
    # simple case
    camera = Camera(center=(0, 0, 0), angles=(90, 0, 0), zoom=1)
    assert np.allclose(camera.view_direction, (0, 1, 0))

    # shouldn't change with zoom
    camera = Camera(center=(0, 0, 0), angles=(90, 0, 0), zoom=10)
    assert np.allclose(camera.view_direction, (0, 1, 0))

    # shouldn't change with center
    camera = Camera(center=(15, 15, 15), angles=(90, 0, 0), zoom=1)
    assert np.allclose(camera.view_direction, (0, 1, 0))


def test_calculate_up_direction_3d():
    """Check that up direction is calculated properly from camera angles."""
    # simple case
    camera = Camera(center=(0, 0, 0), angles=(0, 0, 90), zoom=1)
    assert np.allclose(camera.up_direction, (0, -1, 0))

    # shouldn't change with zoom
    camera = Camera(center=(0, 0, 0), angles=(0, 0, 90), zoom=10)
    assert np.allclose(camera.up_direction, (0, -1, 0))

    # shouldn't change with center
    camera = Camera(center=(15, 15, 15), angles=(0, 0, 90), zoom=1)
    assert np.allclose(camera.up_direction, (0, -1, 0))

    # more complex case with order dependent Euler angles
    camera = Camera(center=(0, 0, 0), angles=(10, 20, 30), zoom=1)
    assert np.allclose(camera.up_direction, (0.88, -0.44, 0.16), atol=0.01)


def test_set_view_direction_3d():
    """Check that view direction can be set properly."""
    # simple case
    camera = Camera(center=(0, 0, 0), angles=(0, 0, 0), zoom=1)
    camera.set_view_direction(view_direction=(1, 0, 0))
    assert np.allclose(camera.view_direction, (1, 0, 0))
    assert np.allclose(camera.angles, (0, 0, 90))

    # case with ordering and up direction setting
    view_direction = np.array([1, 2, 3], dtype=float)
    view_direction /= np.linalg.norm(view_direction)
    camera.set_view_direction(view_direction=view_direction)
    assert np.allclose(camera.view_direction, view_direction)
    assert np.allclose(camera.angles, (58.1, -53.3, 26.6), atol=0.1)


def test_calculate_view_direction_nd():
    """Check that nD view direction is calculated properly."""
    camera = Camera(center=(0, 0, 0), angles=(90, 0, 0), zoom=1)

    # should return none if ndim == 2
    view_direction = camera.calculate_nd_view_direction(
        ndim=2, dims_displayed=[0, 1]
    )
    assert view_direction is None

    # should return 3d if ndim == 3
    view_direction = camera.calculate_nd_view_direction(
        ndim=3, dims_displayed=[0, 1, 2]
    )
    assert len(view_direction) == 3
    assert np.allclose(view_direction, (0, 1, 0))

    # should return nD with 3d embedded in nD if ndim > 3
    view_direction = camera.calculate_nd_view_direction(
        ndim=5, dims_displayed=[0, 2, 4]
    )
    assert len(view_direction) == 5
    assert np.allclose(view_direction[[0, 2, 4]], (0, 1, 0))



import numpy as np
import pytest
from numpy.testing import assert_array_equal
from skimage.util import img_as_ubyte

from napari.layers.utils.layer_utils import convert_to_uint8


@pytest.mark.filterwarnings('ignore:Downcasting uint:UserWarning:skimage')
@pytest.mark.parametrize('dtype', [np.uint8, np.uint16, np.uint32, np.uint64])
def test_uint(dtype):
    data = np.arange(50, dtype=dtype)
    data_scaled = data * 256 ** (data.dtype.itemsize - 1)
    assert convert_to_uint8(data_scaled).dtype == np.uint8
    assert_array_equal(data, convert_to_uint8(data_scaled))
    assert_array_equal(img_as_ubyte(data), convert_to_uint8(data))
    assert_array_equal(
        img_as_ubyte(data_scaled), convert_to_uint8(data_scaled)
    )


@pytest.mark.filterwarnings('ignore:Downcasting int:UserWarning:skimage')
@pytest.mark.parametrize('dtype', [np.int8, np.int16, np.int32, np.int64])
def test_int(dtype):
    data = np.arange(50, dtype=dtype)
    data_scaled = data * 256 ** (data.dtype.itemsize - 1)
    assert convert_to_uint8(data).dtype == np.uint8
    assert convert_to_uint8(data_scaled).dtype == np.uint8
    assert_array_equal(img_as_ubyte(data), convert_to_uint8(data))
    assert_array_equal(2 * data, convert_to_uint8(data_scaled))
    assert_array_equal(
        img_as_ubyte(data_scaled), convert_to_uint8(data_scaled)
    )
    assert_array_equal(img_as_ubyte(data - 10), convert_to_uint8(data - 10))
    assert_array_equal(
        img_as_ubyte(data_scaled - 10), convert_to_uint8(data_scaled - 10)
    )


@pytest.mark.parametrize('dtype', [np.float64, np.float32, float])
def test_float(dtype):
    data = np.linspace(0, 0.5, 128, dtype=dtype, endpoint=False)
    res = np.arange(128, dtype=np.uint8)
    assert convert_to_uint8(data).dtype == np.uint8
    assert_array_equal(convert_to_uint8(data), res)
    data = np.linspace(0, 1, 256, dtype=dtype)
    res = np.arange(256, dtype=np.uint8)
    assert_array_equal(convert_to_uint8(data), res)
    assert_array_equal(img_as_ubyte(data), convert_to_uint8(data))
    assert_array_equal(img_as_ubyte(data - 0.5), convert_to_uint8(data - 0.5))


def test_bool():
    data = np.zeros((10, 10), dtype=bool)
    data[2:-2, 2:-2] = 1
    converted = convert_to_uint8(data)
    assert converted.dtype == np.uint8
    assert_array_equal(img_as_ubyte(data), converted)



from enum import auto

from napari.utils.misc import StringEnum
from napari.utils.translations import trans


class Shading(StringEnum):
    """Shading: Shading mode for the surface.

    Selects a preset shading mode in vispy that determines how
    color is computed in the scene.
    See also: https://www.khronos.org/registry/OpenGL-Refpages/gl2.1/xhtml/glShadeModel.xml
            Shading.NONE
                Computed color is interpreted as input color, unaffected by
                lighting. Corresponds to shading='none'.
            Shading.FLAT
                Computed colours are the color at a specific vertex for each
                primitive in the mesh. Corresponds to shading='flat'.
            Shading.SMOOTH
                Computed colors are interpolated between vertices for each
                primitive in the mesh. Corresponds to shading='smooth'
    """

    NONE = auto()
    FLAT = auto()
    SMOOTH = auto()


SHADING_TRANSLATION = {
    trans._('none'): Shading.NONE,
    trans._('flat'): Shading.FLAT,
    trans._('smooth'): Shading.SMOOTH,
}



from napari.components.overlays.axes import AxesOverlay
from napari.components.overlays.base import (
    CanvasOverlay,
    Overlay,
    SceneOverlay,
)
from napari.components.overlays.bounding_box import BoundingBoxOverlay
from napari.components.overlays.brush_circle import BrushCircleOverlay
from napari.components.overlays.interaction_box import (
    SelectionBoxOverlay,
    TransformBoxOverlay,
)
from napari.components.overlays.labels_polygon import LabelsPolygonOverlay
from napari.components.overlays.scale_bar import ScaleBarOverlay
from napari.components.overlays.text import TextOverlay

__all__ = [
    'AxesOverlay',
    'Overlay',
    'CanvasOverlay',
    'BoundingBoxOverlay',
    'SelectionBoxOverlay',
    'TransformBoxOverlay',
    'LabelsPolygonOverlay',
    'ScaleBarOverlay',
    'SceneOverlay',
    'TextOverlay',
    'BrushCircleOverlay',
]



import itertools
import sys
from unittest.mock import patch

import pyautogui
import pytest
from qtpy.QtCore import QPoint, Qt
from qtpy.QtWidgets import QAbstractItemDelegate, QApplication, QMessageBox

from napari._qt.widgets.qt_keyboard_settings import ShortcutEditor, WarnPopup
from napari._tests.utils import skip_local_focus, skip_on_mac_ci
from napari.settings import get_settings
from napari.utils.action_manager import action_manager
from napari.utils.interactions import KEY_SYMBOLS
from napari.utils.key_bindings import KeyBinding

META_CONTROL_KEY = Qt.KeyboardModifier.ControlModifier
if sys.platform == 'darwin':
    META_CONTROL_KEY = Qt.KeyboardModifier.MetaModifier


@pytest.fixture
def shortcut_editor_widget(qtbot):
    # Always reset shortcuts (settings and action manager)
    get_settings().shortcuts.reset()
    for (
        action,
        shortcuts,
    ) in get_settings().shortcuts.shortcuts.items():
        action_manager.unbind_shortcut(action)
        for shortcut in shortcuts:
            action_manager.bind_shortcut(action, shortcut)

    def _shortcut_editor_widget(**kwargs):
        widget = ShortcutEditor(**kwargs)
        widget._reset_shortcuts()
        widget.show()
        qtbot.addWidget(widget)

        return widget

    return _shortcut_editor_widget


def test_shortcut_editor_defaults(
    shortcut_editor_widget,
):
    shortcut_editor_widget()


def test_potentially_conflicting_actions(shortcut_editor_widget):
    widget = shortcut_editor_widget()
    assert widget.layer_combo_box.currentText() == widget.VIEWER_KEYBINDINGS
    actions1 = widget._get_potential_conflicting_actions()
    expected_actions1 = []
    for group, keybindings in widget.key_bindings_strs.items():
        expected_actions1.extend(
            zip(itertools.repeat(group), keybindings.items())
        )
    assert actions1 == expected_actions1
    widget.layer_combo_box.setCurrentText('Labels layer')
    actions2 = widget._get_potential_conflicting_actions()
    expected_actions2 = list(
        zip(
            itertools.repeat('Labels layer'),
            widget.key_bindings_strs['Labels layer'].items(),
        )
    )
    expected_actions2.extend(
        zip(
            itertools.repeat(widget.VIEWER_KEYBINDINGS),
            widget.key_bindings_strs[widget.VIEWER_KEYBINDINGS].items(),
        )
    )
    assert actions2 == expected_actions2


def test_mark_conflicts(shortcut_editor_widget, qtbot):
    widget = shortcut_editor_widget()
    ctrl_keybinding = KeyBinding.from_str('Ctrl')
    u_keybinding = KeyBinding.from_str('U')
    act = widget._table.item(0, widget._action_col).text()

    # Add check for initial/default keybinding (first shortcuts column) and
    # added one (second shortcuts column)
    assert action_manager._shortcuts[act][0] == ctrl_keybinding
    widget._table.item(0, widget._shortcut_col2).setText(str(u_keybinding))
    assert action_manager._shortcuts[act][1] == str(u_keybinding)

    # Check conflicts detection using `KeyBindingLike` params
    # (`KeyBinding`, `str` and `int` representations of a shortcut)
    with patch.object(WarnPopup, 'exec_') as mock:
        assert not widget._mark_conflicts(ctrl_keybinding, 1)
        assert mock.called
    with patch.object(WarnPopup, 'exec_') as mock:
        assert not widget._mark_conflicts(str(ctrl_keybinding), 1)
        assert mock.called
    with patch.object(WarnPopup, 'exec_') as mock:
        assert not widget._mark_conflicts(int(ctrl_keybinding), 1)
        assert mock.called

    with patch.object(WarnPopup, 'exec_') as mock:
        assert not widget._mark_conflicts(u_keybinding, 1)
        assert mock.called
    with patch.object(WarnPopup, 'exec_') as mock:
        assert not widget._mark_conflicts(str(u_keybinding), 1)
        assert mock.called

    # Check no conflicts are found using `KeyBindingLike` params
    # (`KeyBinding`, `str` and `int` representations of a shortcut)
    # "H" is arbitrary chosen and on conflict with existing shortcut should be changed
    h_keybinding = KeyBinding.from_str('H')
    assert widget._mark_conflicts(h_keybinding, 1)
    assert widget._mark_conflicts(str(h_keybinding), 1)
    assert widget._mark_conflicts(int(h_keybinding), 1)
    qtbot.add_widget(widget._warn_dialog)


def test_restore_defaults(shortcut_editor_widget):
    widget = shortcut_editor_widget()
    shortcut = widget._table.item(0, widget._shortcut_col).text()
    assert shortcut == KEY_SYMBOLS['Ctrl']
    widget._table.item(0, widget._shortcut_col).setText('H')
    shortcut = widget._table.item(0, widget._shortcut_col).text()
    assert shortcut == 'H'
    with patch(
        'napari._qt.widgets.qt_keyboard_settings.QMessageBox.question'
    ) as mock:
        mock.return_value = QMessageBox.RestoreDefaults
        widget._restore_button.click()
        assert mock.called
    shortcut = widget._table.item(0, widget._shortcut_col).text()
    assert shortcut == KEY_SYMBOLS['Ctrl']


@skip_local_focus
@pytest.mark.parametrize(
    ('key', 'modifier', 'key_symbols'),
    [
        (
            Qt.Key.Key_U,
            META_CONTROL_KEY,
            [KEY_SYMBOLS['Ctrl'], 'U'],
        ),
        (
            Qt.Key.Key_Y,
            META_CONTROL_KEY | Qt.KeyboardModifier.ShiftModifier,
            [KEY_SYMBOLS['Ctrl'], KEY_SYMBOLS['Shift'], 'Y'],
        ),
        (
            Qt.Key.Key_Backspace,
            META_CONTROL_KEY,
            [KEY_SYMBOLS['Ctrl'], KEY_SYMBOLS['Backspace']],
        ),
        (
            Qt.Key.Key_Delete,
            META_CONTROL_KEY | Qt.KeyboardModifier.ShiftModifier,
            [KEY_SYMBOLS['Ctrl'], KEY_SYMBOLS['Shift'], KEY_SYMBOLS['Delete']],
        ),
        (
            Qt.Key.Key_Backspace,
            META_CONTROL_KEY | Qt.KeyboardModifier.ShiftModifier,
            [
                KEY_SYMBOLS['Ctrl'],
                KEY_SYMBOLS['Shift'],
                KEY_SYMBOLS['Backspace'],
            ],
        ),
    ],
)
def test_keybinding_with_modifiers(
    shortcut_editor_widget, qtbot, recwarn, key, modifier, key_symbols
):
    widget = shortcut_editor_widget()
    shortcut = widget._table.item(0, widget._shortcut_col).text()
    assert shortcut == KEY_SYMBOLS['Ctrl']

    x = widget._table.columnViewportPosition(widget._shortcut_col)
    y = widget._table.rowViewportPosition(0)
    item_pos = QPoint(x, y)
    index = widget._table.indexAt(item_pos)
    widget._table.setCurrentIndex(index)
    widget._table.edit(index)
    qtbot.waitUntil(lambda: widget._table.focusWidget() is not None)
    editor = widget._table.focusWidget()
    qtbot.keyPress(editor, key, modifier=modifier)
    widget._table.commitData(editor)
    widget._table.closeEditor(editor, QAbstractItemDelegate.NoHint)

    assert len([warn for warn in recwarn if warn.category is UserWarning]) == 0

    shortcut = widget._table.item(0, widget._shortcut_col).text()
    for key_symbol in key_symbols:
        assert key_symbol in shortcut


@skip_local_focus
@pytest.mark.parametrize(
    ('modifiers', 'key_symbols', 'valid'),
    [
        (
            Qt.KeyboardModifier.ShiftModifier,
            [KEY_SYMBOLS['Shift']],
            True,
        ),
        (
            Qt.KeyboardModifier.AltModifier
            | Qt.KeyboardModifier.ShiftModifier,
            [KEY_SYMBOLS['Ctrl']],
            False,
        ),
    ],
)
def test_keybinding_with_only_modifiers(
    shortcut_editor_widget, qtbot, recwarn, modifiers, key_symbols, valid
):
    widget = shortcut_editor_widget()
    shortcut = widget._table.item(0, widget._shortcut_col).text()
    assert shortcut == KEY_SYMBOLS['Ctrl']

    x = widget._table.columnViewportPosition(widget._shortcut_col)
    y = widget._table.rowViewportPosition(0)
    item_pos = QPoint(x, y)
    index = widget._table.indexAt(item_pos)
    widget._table.setCurrentIndex(index)
    widget._table.edit(index)
    qtbot.waitUntil(lambda: widget._table.focusWidget() is not None)
    editor = widget._table.focusWidget()

    with patch.object(WarnPopup, 'exec_') as mock:
        qtbot.keyPress(editor, Qt.Key_Enter, modifier=modifiers)
        widget._table.commitData(editor)
        widget._table.closeEditor(editor, QAbstractItemDelegate.NoHint)
        if valid:
            assert not mock.called
        else:
            assert mock.called

    assert len([warn for warn in recwarn if warn.category is UserWarning]) == 0

    shortcut = widget._table.item(0, widget._shortcut_col).text()
    for key_symbol in key_symbols:
        assert key_symbol in shortcut


@skip_local_focus
@pytest.mark.parametrize(
    'removal_trigger_key',
    [
        Qt.Key.Key_Delete,
        Qt.Key.Key_Backspace,
    ],
)
def test_remove_shortcut(shortcut_editor_widget, qtbot, removal_trigger_key):
    widget = shortcut_editor_widget()
    shortcut = widget._table.item(0, widget._shortcut_col).text()
    assert shortcut == KEY_SYMBOLS['Ctrl']

    x = widget._table.columnViewportPosition(widget._shortcut_col)
    y = widget._table.rowViewportPosition(0)
    item_pos = QPoint(x, y)
    index = widget._table.indexAt(item_pos)
    widget._table.setCurrentIndex(index)
    widget._table.edit(index)
    qtbot.waitUntil(lambda: widget._table.focusWidget() is not None)
    editor = widget._table.focusWidget()
    qtbot.keyClick(editor, removal_trigger_key)
    qtbot.keyClick(editor, Qt.Key.Key_Enter)
    widget._table.commitData(editor)
    widget._table.closeEditor(editor, QAbstractItemDelegate.NoHint)

    shortcut = widget._table.item(0, widget._shortcut_col).text()
    assert shortcut == ''


@skip_local_focus
@skip_on_mac_ci
@pytest.mark.parametrize(
    ('modifier_key', 'modifiers', 'key_symbols'),
    [
        (
            'shift',
            None,
            [KEY_SYMBOLS['Shift']],
        ),
        (
            'ctrl',
            'shift',
            [KEY_SYMBOLS['Ctrl'], KEY_SYMBOLS['Shift']],
        ),
    ],
)
def test_keybinding_editor_modifier_key_detection(
    shortcut_editor_widget,
    qtbot,
    recwarn,
    modifier_key,
    modifiers,
    key_symbols,
):
    """
    Test modifier keys detection with pyautogui to trigger keyboard events
    from the OS.

    Notes:
        * Skipped on macOS CI due to accessibility permissions not being
          settable on macOS GitHub Actions runners.
        * For this test to pass locally, you need to give the Terminal/iTerm
          application accessibility permissions:
              `System Settings > Privacy & Security > Accessibility`

        See https://github.com/asweigart/pyautogui/issues/247 and
        https://github.com/asweigart/pyautogui/issues/247#issuecomment-437668855
    """
    widget = shortcut_editor_widget()
    shortcut = widget._table.item(0, widget._shortcut_col).text()
    assert shortcut == KEY_SYMBOLS['Ctrl']

    x = widget._table.columnViewportPosition(widget._shortcut_col)
    y = widget._table.rowViewportPosition(0)
    item_pos = QPoint(x, y)
    qtbot.mouseClick(
        widget._table.viewport(), Qt.MouseButton.LeftButton, pos=item_pos
    )
    qtbot.mouseDClick(
        widget._table.viewport(), Qt.MouseButton.LeftButton, pos=item_pos
    )
    qtbot.waitUntil(lambda: QApplication.focusWidget() is not None)

    line_edit = QApplication.focusWidget()
    with pyautogui.hold(modifier_key):
        if modifiers:
            pyautogui.keyDown(modifiers)

        def press_check():
            line_edit.selectAll()
            shortcut = line_edit.selectedText()
            all_pressed = True
            for key_symbol in key_symbols:
                all_pressed &= key_symbol in shortcut
            return all_pressed

        qtbot.waitUntil(lambda: press_check())

        if modifiers:
            pyautogui.keyUp(modifiers)

    def release_check():
        line_edit.selectAll()
        shortcut = line_edit.selectedText()
        return shortcut == ''

    qtbot.waitUntil(lambda: release_check())

    qtbot.keyClick(line_edit, Qt.Key_Escape)
    shortcut = widget._table.item(0, widget._shortcut_col).text()
    assert shortcut == KEY_SYMBOLS['Ctrl']



import pytest

from napari._qt.widgets.qt_size_preview import (
    QtFontSizePreview,
    QtSizeSliderPreviewWidget,
)


@pytest.fixture
def preview_widget(qtbot):
    def _preview_widget(**kwargs):
        widget = QtFontSizePreview(**kwargs)
        widget.show()
        qtbot.addWidget(widget)

        return widget

    return _preview_widget


@pytest.fixture
def font_size_preview_widget(qtbot):
    def _font_size_preview_widget(**kwargs):
        widget = QtSizeSliderPreviewWidget(**kwargs)
        widget.show()
        qtbot.addWidget(widget)

        return widget

    return _font_size_preview_widget


# QtFontSizePreview
# ----------------------------------------------------------------------------
def test_qt_font_size_preview_defaults(preview_widget):
    preview_widget()


def test_qt_font_size_preview_text(preview_widget):
    text = 'Some text'
    widget = preview_widget(text=text)
    assert widget.text() == text

    widget = preview_widget()
    widget.setText(text)
    assert widget.text() == text


# QtSizeSliderPreviewWidget
# ----------------------------------------------------------------------------
def test_qt_size_slider_preview_widget_defaults(font_size_preview_widget):
    font_size_preview_widget()


def test_qt_size_slider_preview_widget_description(font_size_preview_widget):
    description = 'Some text'
    widget = font_size_preview_widget(description=description)
    assert widget.description() == description

    widget = font_size_preview_widget()
    widget.setDescription(description)
    assert widget.description() == description


def test_qt_size_slider_preview_widget_unit(font_size_preview_widget):
    unit = 'EM'
    widget = font_size_preview_widget(unit=unit)
    assert widget.unit() == unit

    widget = font_size_preview_widget()
    widget.setUnit(unit)
    assert widget.unit() == unit


def test_qt_size_slider_preview_widget_preview(font_size_preview_widget):
    preview = 'Some preview'
    widget = font_size_preview_widget(preview_text=preview)
    assert widget.previewText() == preview

    widget = font_size_preview_widget()
    widget.setPreviewText(preview)
    assert widget.previewText() == preview


def test_qt_size_slider_preview_widget_minimum(font_size_preview_widget):
    minimum = 10
    widget = font_size_preview_widget(min_value=minimum)
    assert widget.minimum() == minimum
    assert widget.value() >= minimum

    widget = font_size_preview_widget()
    widget.setMinimum(5)
    assert widget.minimum() == 5
    assert widget._slider.minimum() == 5
    assert widget._slider_min_label.text() == '5'

    widget.setMinimum(20)
    assert widget.minimum() == 20
    assert widget.value() == 20
    assert widget._slider.minimum() == 20
    assert widget._slider_min_label.text() == '20'
    assert widget._lineedit.text() == '20'


def test_qt_size_slider_preview_widget_minimum_invalid(
    font_size_preview_widget,
):
    widget = font_size_preview_widget()

    with pytest.raises(ValueError, match='must be smaller than'):
        widget.setMinimum(60)


def test_qt_size_slider_preview_widget_maximum(font_size_preview_widget):
    maximum = 10
    widget = font_size_preview_widget(max_value=maximum)

    assert widget.maximum() == maximum
    assert widget.value() <= maximum

    widget = font_size_preview_widget()
    widget.setMaximum(20)
    assert widget.maximum() == 20
    assert widget._slider.maximum() == 20
    assert widget._slider_max_label.text() == '20'

    widget.setMaximum(5)
    assert widget.maximum() == 5
    assert widget.value() == 5
    assert widget._slider.maximum() == 5
    assert widget._lineedit.text() == '5'
    assert widget._slider_max_label.text() == '5'


def test_qt_size_slider_preview_widget_maximum_invalid(
    font_size_preview_widget,
):
    widget = font_size_preview_widget()

    with pytest.raises(ValueError, match='must be larger than'):
        widget.setMaximum(-5)


def test_qt_size_slider_preview_widget_value(font_size_preview_widget):
    widget = font_size_preview_widget(value=5)
    assert widget.value() <= 5

    widget = font_size_preview_widget()
    widget.setValue(5)
    assert widget.value() == 5


def test_qt_size_slider_preview_widget_value_invalid(
    qtbot, font_size_preview_widget
):
    widget = font_size_preview_widget()
    widget.setMaximum(50)
    widget.setValue(51)
    assert widget.value() == 50
    assert widget._lineedit.text() == '50'

    widget.setMinimum(5)
    widget.setValue(1)
    assert widget.value() == 5
    assert widget._lineedit.text() == '5'


def test_qt_size_slider_preview_signal(qtbot, font_size_preview_widget):
    widget = font_size_preview_widget()

    with qtbot.waitSignal(widget.valueChanged, timeout=500):
        widget.setValue(7)

    with qtbot.waitSignal(widget.valueChanged, timeout=500):
        widget.setValue(-5)



import pytest

from napari.utils import misc


def test_proxy_fixture_warning(make_napari_viewer_proxy, monkeypatch):
    viewer = make_napari_viewer_proxy()

    monkeypatch.setattr(misc, 'ROOT_DIR', '/some/other/package')
    with pytest.warns(FutureWarning, match='Private attribute access'):
        viewer.window._qt_window


def test_proxy_fixture_thread_error(
    make_napari_viewer_proxy, single_threaded_executor
):
    viewer = make_napari_viewer_proxy()
    future = single_threaded_executor.submit(
        viewer.__setattr__, 'status', 'hi'
    )
    with pytest.raises(RuntimeError, match='Setting attributes'):
        future.result()



import numpy as np
import pytest

from napari.layers import Image
from napari.layers.utils.stack_utils import (
    images_to_stack,
    split_channels,
    stack_to_images,
)
from napari.utils.transforms import Affine


def test_stack_to_images_basic():
    """Test that a 2 channel zcyx stack is split into 2 image layers"""
    data = np.random.randint(0, 100, (10, 2, 128, 128))
    stack = Image(data)
    images = stack_to_images(stack, 1, colormap=None)

    assert isinstance(images, list)
    assert images[0].colormap.name == 'magenta'
    assert len(images) == 2

    for i in images:
        assert type(stack) is type(i)
        assert i.data.shape == (10, 128, 128)


def test_stack_to_images_multiscale():
    """Test that a 3 channel multiscale image returns 3 multiscale images."""
    data = []
    data.append(np.zeros((3, 128, 128)))
    data.append(np.zeros((3, 64, 64)))
    data.append(np.zeros((3, 32, 32)))
    data.append(np.zeros((3, 16, 16)))

    stack = Image(data)
    images = stack_to_images(stack, 0)

    assert len(images) == 3
    assert len(images[0].data) == 4
    assert images[0].data[-1].shape[-1] == 16
    assert images[1].data[-1].shape[-1] == 16
    assert images[2].data[-1].shape[-1] == 16


def test_stack_to_images_rgb():
    """Test 3 channel RGB image (channel axis = -1) into single channels."""
    data = np.random.randint(0, 100, (10, 128, 128, 3))
    stack = Image(data)
    images = stack_to_images(stack, -1, colormap=None)

    assert isinstance(images, list)
    assert len(images) == 3

    for i in images:
        assert type(stack) is type(i)
        assert i.data.shape == (10, 128, 128)
        assert i.scale.shape == (3,)
        assert i.rgb is False


def test_stack_to_images_4_channels():
    """Test 4x128x128 stack is split into 4 channels w/ colormap keyword"""
    data = np.random.randint(0, 100, (4, 128, 128))
    stack = Image(data)
    images = stack_to_images(stack, 0, colormap=['red', 'blue'])

    assert isinstance(images, list)
    assert len(images) == 4
    assert images[-2].colormap.name == 'red'
    for i in images:
        assert type(stack) is type(i)
        assert i.data.shape == (128, 128)


def test_stack_to_images_0_rgb():
    """Split RGB along the first axis (z or t) so the images remain rgb"""
    data = np.random.randint(0, 100, (10, 128, 128, 3))
    stack = Image(data)
    images = stack_to_images(stack, 0, colormap=None)

    assert isinstance(images, list)
    assert len(images) == 10
    for i in images:
        assert i.rgb
        assert type(stack) is type(i)
        assert i.data.shape == (128, 128, 3)


def test_stack_to_images_1_channel():
    """Split when only one channel"""
    data = np.random.randint(0, 100, (10, 1, 128, 128))
    stack = Image(data)
    images = stack_to_images(stack, 1, colormap=['magma'])

    assert isinstance(images, list)
    assert len(images) == 1
    for i in images:
        assert i.rgb is False
        assert type(stack) is type(i)
        assert i.data.shape == (10, 128, 128)


def test_images_to_stack_with_scale():
    """Test that 3-Image list is combined to stack with scale and translate."""
    images = [
        Image(np.random.randint(0, 255, (10, 128, 128))) for _ in range(3)
    ]

    stack = images_to_stack(
        images, 1, colormap='green', scale=(3, 1, 1, 1), translate=(1, 0, 2, 3)
    )

    assert isinstance(stack, Image)
    assert stack.data.shape == (10, 3, 128, 128)
    assert stack.colormap.name == 'green'
    assert list(stack.scale) == [3, 1, 1, 1]
    assert list(stack.translate) == [1, 0, 2, 3]


def test_images_to_stack_none_scale():
    """Test combining images using scale & translate from 1st image in list"""
    images = [
        Image(
            np.random.randint(0, 255, (10, 128, 128)),
            scale=(4, 1, 1),
            translate=(0, -1, 2),
        )
        for _ in range(3)
    ]

    stack = images_to_stack(images, 1, colormap='green')

    assert isinstance(stack, Image)
    assert stack.data.shape == (10, 3, 128, 128)
    assert stack.colormap.name == 'green'
    assert list(stack.scale) == [4, 1, 1, 1]
    assert list(stack.translate) == [0, 0, -1, 2]


@pytest.fixture(
    params=[
        {
            'rgb': None,
            'colormap': None,
            'contrast_limits': None,
            'gamma': 1,
            'interpolation': 'nearest',
            'rendering': 'mip',
            'iso_threshold': 0.5,
            'attenuation': 0.5,
            'name': None,
            'metadata': None,
            'scale': None,
            'translate': None,
            'opacity': 1,
            'blending': None,
            'visible': True,
            'multiscale': None,
            'rotate': None,
            'affine': None,
        },
        {
            'rgb': None,
            'colormap': None,
            'rendering': 'mip',
            'attenuation': 0.5,
            'metadata': None,
            'scale': None,
            'opacity': 1,
            'visible': True,
            'multiscale': None,
        },
        {},
    ],
    ids=['full-kwargs', 'partial-kwargs', 'empty-kwargs'],
)
def kwargs(request):
    return dict(request.param)


def test_split_channels(kwargs):
    """Test split_channels with shape (3,128,128) expecting 3 (128,128)"""
    data = np.zeros((3, 128, 128))
    result_list = split_channels(data, 0, **kwargs)

    assert len(result_list) == 3
    for d, _meta, _ in result_list:
        assert d.shape == (128, 128)


def test_split_channels_multiscale(kwargs):
    """Test split_channels with multiscale expecting List[LayerData]"""
    data = []
    data.append(np.zeros((3, 128, 128)))
    data.append(np.zeros((3, 64, 64)))
    data.append(np.zeros((3, 32, 32)))
    data.append(np.zeros((3, 16, 16)))
    result_list = split_channels(data, 0, **kwargs)

    assert len(result_list) == 3
    for ds, m, _ in result_list:
        assert m['multiscale'] is True
        assert ds[0].shape == (128, 128)
        assert ds[1].shape == (64, 64)
        assert ds[2].shape == (32, 32)
        assert ds[3].shape == (16, 16)


def test_split_channels_blending(kwargs):
    """Test split_channels with shape (3,128,128) expecting 3 (128,128)"""
    kwargs['blending'] = 'translucent'
    data = np.zeros((3, 128, 128))
    result_list = split_channels(data, 0, **kwargs)

    assert len(result_list) == 3
    for d, meta, _ in result_list:
        assert d.shape == (128, 128)
        assert meta['blending'] == 'translucent'


def test_split_channels_missing_keywords():
    data = np.zeros((3, 128, 128))
    result_list = split_channels(data, 0)

    assert len(result_list) == 3
    for chan, layer in enumerate(result_list):
        assert layer[0].shape == (128, 128)
        assert (
            layer[1]['blending'] == 'translucent_no_depth'
            if chan == 0
            else 'additive'
        )


def test_split_channels_affine_nparray(kwargs):
    kwargs['affine'] = np.eye(3)
    data = np.zeros((3, 128, 128))
    result_list = split_channels(data, 0, **kwargs)

    assert len(result_list) == 3
    for d, meta, _ in result_list:
        assert d.shape == (128, 128)
        assert np.array_equal(meta['affine'], np.eye(3))


def test_split_channels_affine_napari(kwargs):
    kwargs['affine'] = Affine(affine_matrix=np.eye(3))
    data = np.zeros((3, 128, 128))
    result_list = split_channels(data, 0, **kwargs)

    assert len(result_list) == 3
    for d, meta, _ in result_list:
        assert d.shape == (128, 128)
        assert np.array_equal(meta['affine'].affine_matrix, np.eye(3))


def test_split_channels_multi_affine_napari(kwargs):
    kwargs['affine'] = [
        Affine(scale=[1, 1]),
        Affine(scale=[2, 2]),
        Affine(scale=[3, 3]),
    ]

    data = np.zeros((3, 128, 128))
    result_list = split_channels(data, 0, **kwargs)

    assert len(result_list) == 3
    for idx, result_data in enumerate(result_list):
        d, meta, _ = result_data
        assert d.shape == (128, 128)
        assert np.array_equal(
            meta['affine'].affine_matrix,
            Affine(scale=[idx + 1, idx + 1]).affine_matrix,
        )



import napari


def test_top_level_availability(make_napari_viewer):
    """Current viewer should be available at napari.current_viewer."""
    viewer = make_napari_viewer()
    assert viewer == napari.current_viewer()



import struct

import numpy as np
import pytest
import tifffile
from imageio.v3 import imread

from napari.utils.io import imsave


@pytest.mark.parametrize(
    'image_file', ['image', 'image.png', 'image.tif', 'image.bmp']
)
def test_imsave(tmp_path, image_file):
    """Save data to image file."""
    # create image data
    np.random.seed(0)
    data = np.random.randint(20, size=(10, 15), dtype=np.ubyte)
    image_file_path = tmp_path / image_file
    assert not image_file_path.is_file()

    # create image and assert image file creation
    imsave(str(image_file_path), data)
    assert image_file_path.is_file()

    # check image content
    img_to_array = imread(str(image_file_path))
    assert np.equal(data, img_to_array).all()


def test_imsave_bool_tiff(tmp_path):
    """Test saving a boolean array to a tiff file."""
    np.random.seed(0)
    data = np.random.randint(low=0, high=2, size=(10, 15), dtype=bool)
    image_file_path = tmp_path / 'bool_image.tif'
    assert not image_file_path.is_file()

    # create image and assert image file creation
    imsave(str(image_file_path), data)
    assert image_file_path.is_file()

    # check image content
    img_to_array = imread(str(image_file_path))
    assert np.equal(data, img_to_array).all()


@pytest.mark.parametrize(
    'image_file', ['image', 'image.png', 'image.tif', 'image.bmp']
)
def test_imsave_float(tmp_path, image_file):
    """Test saving float image data."""
    # create image data
    np.random.seed(0)
    data = np.random.random((10, 15))
    image_file_path = tmp_path / image_file
    assert not image_file_path.is_file()

    # create image
    imsave(str(image_file_path), data)
    # only TIF can store float
    if image_file.endswith('.tif'):
        assert image_file_path.is_file()
        img_to_array = imread(str(image_file_path))
        assert np.equal(data, img_to_array).all()
    # if no EXT provided, for float data should save as .tif
    elif image_file == 'image':
        assert image_file_path.with_suffix('.tif').is_file()
        img_to_array = imread(str(image_file_path.with_suffix('.tif')))
        assert np.equal(data, img_to_array).all()

    else:
        assert not image_file_path.is_file()


def test_imsave_large_file(monkeypatch, tmp_path):
    old_write = tifffile.imwrite

    def raise_no_bigtiff(*args, **kwargs):
        if 'bigtiff' not in kwargs:
            raise struct.error
        old_write(*args, **kwargs)

    monkeypatch.setattr(tifffile, 'imwrite', raise_no_bigtiff)

    data = np.random.randint(
        low=0, high=2**16, size=(128, 4096, 4096), dtype='uint16'
    )  # 4GB size

    # create image and assert image file creation
    image_path = str(tmp_path / 'data.tif')
    imsave(image_path, data)
    with tifffile.TiffFile(image_path) as tiff:
        assert tiff.is_bigtiff



from unittest.mock import MagicMock

import pytest

from napari._qt.threads.status_checker import StatusChecker
from napari.components import ViewerModel


@pytest.mark.usefixtures('qapp')
def test_create():
    StatusChecker(ViewerModel())


@pytest.mark.usefixtures('qapp')
def test_no_emmit_no_ref(monkeypatch):
    """Calling calculate_status should not emit after viewer is deleted."""
    model = ViewerModel()
    status_checker = StatusChecker(model)
    monkeypatch.setattr(
        status_checker,
        'status_and_tooltip_changed',
        MagicMock(side_effect=RuntimeError('Should not emit')),
    )
    del model
    status_checker.calculate_status()


def test_terminate_no_ref(monkeypatch):
    """Test that the thread terminates when the viewer is garbage collected."""
    model = ViewerModel()
    status_checker = StatusChecker(model)
    del model
    status_checker.run()
    assert not status_checker._terminate


def test_waiting_on_no_request(monkeypatch, qtbot):
    def _check_status(value):
        return value == ('Ready', '')

    model = ViewerModel()
    model.mouse_over_canvas = True
    status_checker = StatusChecker(model)
    status_checker.start()
    with qtbot.waitSignal(
        status_checker.status_and_tooltip_changed,
        timeout=1000,
        check_params_cb=_check_status,
    ):
        status_checker.trigger_status_update()
    status_checker.terminate()

    qtbot.wait_until(lambda: status_checker.isFinished())



import re
from collections.abc import Iterable, Iterator
from functools import lru_cache
from itertools import product
from pathlib import Path
from typing import Optional, Union

from napari.utils._appdirs import user_cache_dir
from napari.utils.translations import trans

LOADING_GIF_PATH = str((Path(__file__).parent / 'loading.gif').resolve())
ICON_PATH = (Path(__file__).parent / 'icons').resolve()
ICONS = {x.stem: str(x) for x in ICON_PATH.iterdir() if x.suffix == '.svg'}
PLUGIN_FILE_NAME = 'plugin.txt'


def get_icon_path(name: str) -> str:
    """Return path to an SVG in the theme icons."""
    if name not in ICONS:
        raise ValueError(
            trans._(
                'unrecognized icon name: {name!r}. Known names: {icons}',
                deferred=True,
                name=name,
                icons=set(ICONS),
            )
        )
    return ICONS[name]


svg_elem = re.compile(r'(<svg[^>]*>)')
svg_style = """<style type="text/css">
path {{fill: {0}; opacity: {1};}}
polygon {{fill: {0}; opacity: {1};}}
circle {{fill: {0}; opacity: {1};}}
rect {{fill: {0}; opacity: {1};}}
</style>"""


@lru_cache
def get_raw_svg(path: str) -> str:
    """Get and cache SVG XML."""
    return Path(path).read_text()


@lru_cache
def get_colorized_svg(
    path_or_xml: Union[str, Path], color: Optional[str] = None, opacity=1.0
) -> str:
    """Return a colorized version of the SVG XML at ``path``.

    Raises
    ------
    ValueError
        If the path exists but does not contain valid SVG data.
    """
    path_or_xml = str(path_or_xml)
    xml = path_or_xml if '</svg>' in path_or_xml else get_raw_svg(path_or_xml)
    if not color:
        return xml

    if not svg_elem.search(xml):
        raise ValueError(
            trans._(
                'Could not detect svg tag in {path_or_xml!r}',
                deferred=True,
                path_or_xml=path_or_xml,
            )
        )
    # use regex to find the svg tag and insert css right after
    # (the '\\1' syntax includes the matched tag in the output)
    return svg_elem.sub(f'\\1{svg_style.format(color, opacity)}', xml)


def generate_colorized_svgs(
    svg_paths: Iterable[Union[str, Path]],
    colors: Iterable[Union[str, tuple[str, str]]],
    opacities: Iterable[float] = (1.0,),
    theme_override: Optional[dict[str, str]] = None,
) -> Iterator[tuple[str, str]]:
    """Helper function to generate colorized SVGs.

    This is a generator that yields tuples of ``(alias, icon_xml)`` for every
    combination (Cartesian product) of `svg_path`, `color`, and `opacity`
    provided. It can be used as input to :func:`_temporary_qrc_file`.

    Parameters
    ----------
    svg_paths : Iterable[Union[str, Path]]
        An iterable of paths to svg files
    colors : Iterable[Union[str, Tuple[str, str]]]
        An iterable of colors.  Every icon will be generated in every color. If
        a `color` item is a string, it should be valid svg color style. Items
        may also be a 2-tuple of strings, in which case the first item should
        be an available theme name
        (:func:`~napari.utils.theme.available_themes`), and the second item
        should be a key in the theme (:func:`~napari.utils.theme.get_theme`),
    opacities : Iterable[float], optional
        An iterable of opacities to generate, by default (1.0,) Opacities less
        than one can be accessed in qss with the opacity as a percentage
        suffix, e.g.: ``my_svg_50.svg`` for opacity 0.5.
    theme_override : Optional[Dict[str, str]], optional
        When one of the `colors` is a theme ``(name, key)`` tuple,
        `theme_override` may be used to override the `key` for a specific icon
        name in `svg_paths`.  For example ``{'exclamation': 'warning'}``, would
        use the theme "warning" color for any icon named "exclamation.svg" by
        default `None`

    Yields
    ------
    (alias, xml) : Iterator[Tuple[str, str]]
        `alias` is the name that will used to access the icon in the Qt
        Resource system (such as QSS), and `xml` is the *raw* colorzied SVG
        text (as read from a file, perhaps pre-colored using one of the below
        functions).
    """

    # mapping of svg_stem to theme_key
    theme_override = theme_override or {}

    ALIAS_T = '{color}/{svg_stem}{opacity}.svg'

    for color, path, op in product(colors, svg_paths, opacities):
        clrkey = color
        svg_stem = Path(path).stem
        if isinstance(color, tuple):
            from napari.utils.theme import get_theme

            clrkey, theme_key = color
            theme_key = theme_override.get(svg_stem, theme_key)
            color = getattr(get_theme(clrkey), theme_key).as_hex()
            # convert color to string to fit get_colorized_svg signature

        op_key = '' if op == 1 else f'_{op * 100:.0f}'
        alias = ALIAS_T.format(color=clrkey, svg_stem=svg_stem, opacity=op_key)
        yield alias, get_colorized_svg(path, color, op)


def write_colorized_svgs(
    dest: Union[str, Path],
    svg_paths: Iterable[Union[str, Path]],
    colors: Iterable[Union[str, tuple[str, str]]],
    opacities: Iterable[float] = (1.0,),
    theme_override: Optional[dict[str, str]] = None,
):
    dest = Path(dest)
    dest.mkdir(parents=True, exist_ok=True)
    svgs = generate_colorized_svgs(
        svg_paths=svg_paths,
        colors=colors,
        opacities=opacities,
        theme_override=theme_override,
    )

    for alias, svg in svgs:
        (dest / Path(alias).name).write_text(svg)


def _theme_path(theme_name: str) -> Path:
    return Path(user_cache_dir()) / '_themes' / theme_name


def build_theme_svgs(theme_name: str, source) -> str:
    out = _theme_path(theme_name)
    write_colorized_svgs(
        out,
        svg_paths=ICONS.values(),
        colors=[(theme_name, 'icon')],
        opacities=(0.5, 1),
        theme_override={
            'warning': 'warning',
            'error': 'error',
            'logo_silhouette': 'background',
        },
    )
    with (out / PLUGIN_FILE_NAME).open('w') as f:
        f.write(source)
    return str(out)



import numpy as np
from vispy.scene import ArcballCamera, BaseCamera, PanZoomCamera

from napari._vispy.utils.quaternion import quaternion2euler_degrees


class VispyCamera:
    """Vipsy camera for both 2D and 3D rendering.

    Parameters
    ----------
    view : vispy.scene.widgets.viewbox.ViewBox
        Viewbox for current scene.
    camera : napari.components.Camera
        napari camera model.
    dims : napari.components.Dims
        napari dims model.
    """

    def __init__(self, view, camera, dims) -> None:
        self._view = view
        self._camera = camera
        self._dims = dims

        # Create 2D camera
        self._2D_camera = MouseToggledPanZoomCamera(aspect=1)
        # flip y-axis to have correct alignment
        self._2D_camera.flip = (0, 1, 0)
        self._2D_camera.viewbox_key_event = viewbox_key_event

        # Create 3D camera
        self._3D_camera = MouseToggledArcballCamera(fov=0)
        self._3D_camera.viewbox_key_event = viewbox_key_event

        # Set 2D camera by default
        self._view.camera = self._2D_camera

        self._dims.events.ndisplay.connect(
            self._on_ndisplay_change, position='first'
        )

        self._camera.events.center.connect(self._on_center_change)
        self._camera.events.zoom.connect(self._on_zoom_change)
        self._camera.events.angles.connect(self._on_angles_change)
        self._camera.events.perspective.connect(self._on_perspective_change)
        self._camera.events.mouse_pan.connect(self._on_mouse_toggles_change)
        self._camera.events.mouse_zoom.connect(self._on_mouse_toggles_change)

        self._on_ndisplay_change()

    @property
    def angles(self):
        """3-tuple: Euler angles of camera in 3D viewing, in degrees.
        Note that angles might be different than the ones that might have generated the quaternion.
        """

        if self._view.camera == self._3D_camera:
            # Do conversion from quaternion representation to euler angles
            angles = quaternion2euler_degrees(self._view.camera._quaternion)
        else:
            angles = (0, 0, 90)
        return angles

    @angles.setter
    def angles(self, angles):
        if self.angles == tuple(angles):
            return

        # Only update angles if current camera is 3D camera
        if self._view.camera == self._3D_camera:
            # Create and set quaternion
            quat = self._view.camera._quaternion.create_from_euler_angles(
                *angles,
                degrees=True,
            )
            self._view.camera._quaternion = quat
            self._view.camera.view_changed()

    @property
    def center(self):
        """tuple: Center point of camera view for 2D or 3D viewing."""
        if self._view.camera == self._3D_camera:
            center = tuple(self._view.camera.center)
        else:
            # in 2D, we arbitrarily choose 0.0 as the center in z
            center = (*self._view.camera.center[:2], 0.0)
        # switch from VisPy xyz ordering to NumPy prc ordering
        return center[::-1]

    @center.setter
    def center(self, center):
        if self.center == tuple(center):
            return
        self._view.camera.center = center[::-1]
        self._view.camera.view_changed()

    @property
    def zoom(self):
        """float: Scale from canvas pixels to world pixels."""
        canvas_size = np.array(self._view.canvas.size)
        if self._view.camera == self._3D_camera:
            # For fov = 0.0 normalize scale factor by canvas size to get scale factor.
            # Note that the scaling is stored in the `_projection` property of the
            # camera which is updated in vispy here
            # https://github.com/vispy/vispy/blob/v0.6.5/vispy/scene/cameras/perspective.py#L301-L313
            scale = self._view.camera.scale_factor
        else:
            scale = np.array(
                [self._view.camera.rect.width, self._view.camera.rect.height]
            )
            scale[np.isclose(scale, 0)] = 1  # fix for #2875
        zoom = np.min(canvas_size / scale)
        return zoom

    @zoom.setter
    def zoom(self, zoom):
        if self.zoom == zoom:
            return
        scale = np.array(self._view.canvas.size) / zoom
        if self._view.camera == self._3D_camera:
            self._view.camera.scale_factor = np.min(scale)
        else:
            # Set view rectangle, as left, right, width, height
            corner = np.subtract(self._view.camera.center[:2], scale / 2)
            self._view.camera.rect = tuple(corner) + tuple(scale)

    @property
    def perspective(self):
        """Field of view of camera (only visible in 3D mode)."""
        return self._3D_camera.fov

    @perspective.setter
    def perspective(self, perspective):
        if self.perspective == perspective:
            return
        self._3D_camera.fov = perspective
        self._view.camera.view_changed()

    @property
    def mouse_zoom(self) -> bool:
        return self._view.camera.mouse_zoom

    @mouse_zoom.setter
    def mouse_zoom(self, mouse_zoom: bool):
        self._view.camera.mouse_zoom = mouse_zoom

    @property
    def mouse_pan(self) -> bool:
        return self._view.camera.mouse_pan

    @mouse_pan.setter
    def mouse_pan(self, mouse_pan: bool):
        self._view.camera.mouse_pan = mouse_pan

    def _on_ndisplay_change(self):
        if self._dims.ndisplay == 3:
            self._view.camera = self._3D_camera
        else:
            self._view.camera = self._2D_camera

        self._on_mouse_toggles_change()
        self._on_center_change()
        self._on_zoom_change()
        self._on_angles_change()

    def _on_mouse_toggles_change(self):
        self.mouse_pan = self._camera.mouse_pan
        self.mouse_zoom = self._camera.mouse_zoom

    def _on_center_change(self):
        self.center = self._camera.center[-self._dims.ndisplay :]

    def _on_zoom_change(self):
        self.zoom = self._camera.zoom

    def _on_perspective_change(self):
        self.perspective = self._camera.perspective

    def _on_angles_change(self):
        self.angles = self._camera.angles

    def on_draw(self, _event):
        """Called whenever the canvas is drawn.

        Update camera model angles, center, and zoom.
        """
        with self._camera.events.angles.blocker(self._on_angles_change):
            self._camera.angles = self.angles
        with self._camera.events.center.blocker(self._on_center_change):
            self._camera.center = self.center
        with self._camera.events.zoom.blocker(self._on_zoom_change):
            self._camera.zoom = self.zoom
        with self._camera.events.perspective.blocker(
            self._on_perspective_change
        ):
            self._camera.perspective = self.perspective


def viewbox_key_event(event):
    """ViewBox key event handler.

    Parameters
    ----------
    event : vispy.util.event.Event
        The vispy event that triggered this method.
    """
    return


def add_mouse_pan_zoom_toggles(
    vispy_camera_cls: type[BaseCamera],
) -> type[BaseCamera]:
    """Add separate mouse pan and mouse zoom toggles to VisPy.

    By default, VisPy uses an ``interactive`` toggle that turns *both*
    panning and zooming on and off. This decorator adds separate toggles,
    ``mouse_pan`` and ``mouse_zoom``, to enable controlling them
    separately.

    Parameters
    ----------
    vispy_camera_cls : Type[vispy.scene.cameras.BaseCamera]
        A VisPy camera class to decorate.

    Returns
    -------
    A decorated VisPy camera class.
    """

    class _vispy_camera_cls(vispy_camera_cls):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.mouse_pan = True
            self.mouse_zoom = True

        def viewbox_mouse_event(self, event):
            if (
                self.mouse_zoom
                and event.type in ('mouse_wheel', 'gesture_zoom')
                or self.mouse_pan
                and event.type
                in ('mouse_move', 'mouse_press', 'mouse_release')
            ):
                super().viewbox_mouse_event(event)
            else:
                event.handled = False

    return _vispy_camera_cls


MouseToggledPanZoomCamera = add_mouse_pan_zoom_toggles(PanZoomCamera)
MouseToggledArcballCamera = add_mouse_pan_zoom_toggles(ArcballCamera)



from napari._app_model.constants._menus import MenuId
from napari._app_model.utils import to_id_key


def assert_empty_keys_in_context(viewer):
    context = viewer.window._qt_viewer._layers.model().sourceModel()._root._ctx
    for menu_id in MenuId.contributables():
        context_key = f'{to_id_key(menu_id)}_empty'
        assert context_key in context


def test_menu_viewer_relaunch(make_napari_viewer):
    viewer = make_napari_viewer()
    assert_empty_keys_in_context(viewer)
    viewer.close()

    viewer2 = make_napari_viewer()
    # prior to #7106, this would fail
    assert_empty_keys_in_context(viewer2)
    viewer2.close()

    # prior to #7106, creating this viewer would error
    make_napari_viewer()



from vispy.scene.visuals import Text

from napari._vispy.overlays.base import ViewerOverlayMixin, VispyCanvasOverlay
from napari.components._viewer_constants import CanvasPosition


class VispyTextOverlay(ViewerOverlayMixin, VispyCanvasOverlay):
    """Text overlay."""

    def __init__(self, *, viewer, overlay, parent=None) -> None:
        super().__init__(
            node=Text(pos=(0, 0)),
            viewer=viewer,
            overlay=overlay,
            parent=parent,
        )

        self.node.font_size = self.overlay.font_size
        self.node.anchors = ('left', 'top')

        self.overlay.events.text.connect(self._on_text_change)
        self.overlay.events.color.connect(self._on_color_change)
        self.overlay.events.font_size.connect(self._on_font_size_change)

        self.reset()

    def _on_text_change(self):
        self.node.text = self.overlay.text

    def _on_color_change(self):
        self.node.color = self.overlay.color

    def _on_font_size_change(self):
        self.node.font_size = self.overlay.font_size

    def _on_position_change(self, event=None):
        super()._on_position_change()
        position = self.overlay.position

        if position == CanvasPosition.TOP_LEFT:
            anchors = ('left', 'bottom')
        elif position == CanvasPosition.TOP_RIGHT:
            anchors = ('right', 'bottom')
        elif position == CanvasPosition.TOP_CENTER:
            anchors = ('center', 'bottom')
        elif position == CanvasPosition.BOTTOM_RIGHT:
            anchors = ('right', 'top')
        elif position == CanvasPosition.BOTTOM_LEFT:
            anchors = ('left', 'top')
        elif position == CanvasPosition.BOTTOM_CENTER:
            anchors = ('center', 'top')

        self.node.anchors = anchors

    def reset(self):
        super().reset()
        self._on_text_change()
        self._on_color_change()
        self._on_font_size_change()



from typing import (
    Any,
    Literal,
    Optional,
    Protocol,
    Union,
    runtime_checkable,
)

import numpy as np

from napari._pydantic_compat import Field, parse_obj_as, validator
from napari.layers.utils.color_transformations import ColorType
from napari.layers.utils.style_encoding import (
    StyleEncoding,
    _ConstantStyleEncoding,
    _DerivedStyleEncoding,
    _ManualStyleEncoding,
)
from napari.utils import Colormap
from napari.utils.color import ColorArray, ColorValue
from napari.utils.colormaps import ValidColormapArg, ensure_colormap
from napari.utils.colormaps.categorical_colormap import CategoricalColormap
from napari.utils.translations import trans

"""The default color to use, which may also be used a safe fallback color."""
DEFAULT_COLOR = ColorValue.validate('cyan')


@runtime_checkable
class ColorEncoding(StyleEncoding[ColorValue, ColorArray], Protocol):
    """Encodes colors from features."""

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(
        cls, value: Union['ColorEncoding', dict, str, ColorType]
    ) -> 'ColorEncoding':
        """Validates and coerces a value to a ColorEncoding.

        Parameters
        ----------
        value : ColorEncodingArgument
            The value to validate and coerce.
            If this is already a ColorEncoding, it is returned as is.
            If this is a dict, then it should represent one of the built-in color encodings.
            If this a string, then a DirectColorEncoding is returned.
            If this a single color, a ConstantColorEncoding is returned.
            If this is a sequence of colors, a ManualColorEncoding is returned.

        Returns
        -------
        ColorEncoding

        Raises
        ------
        TypeError
            If the value is not a supported type.
        ValidationError
            If the value cannot be parsed into a ColorEncoding.
        """
        if isinstance(value, ColorEncoding):
            return value
        if isinstance(value, dict):
            return parse_obj_as(
                Union[
                    ConstantColorEncoding,
                    ManualColorEncoding,
                    DirectColorEncoding,
                    NominalColorEncoding,
                    QuantitativeColorEncoding,
                ],
                value,
            )
        try:
            color_array = ColorArray.validate(value)
        except (ValueError, AttributeError, KeyError) as e:
            raise TypeError(
                trans._(
                    'value should be a ColorEncoding, a dict, a color, or a sequence of colors',
                    deferred=True,
                )
            ) from e
        if color_array.shape[0] == 1:
            return ConstantColorEncoding(constant=value)
        return ManualColorEncoding(array=color_array, default=DEFAULT_COLOR)


class ConstantColorEncoding(_ConstantStyleEncoding[ColorValue, ColorArray]):
    """Encodes color values from a single constant color.

    Attributes
    ----------
    constant : ColorValue
        The constant color RGBA value.
    """

    encoding_type: Literal['ConstantColorEncoding'] = 'ConstantColorEncoding'
    constant: ColorValue


class ManualColorEncoding(_ManualStyleEncoding[ColorValue, ColorArray]):
    """Encodes color values manually in an array attribute.

    Attributes
    ----------
    array : ColorArray
        The array of color values. Can be written to directly to make
        persistent updates.
    default : ColorValue
        The default color value.
    """

    encoding_type: Literal['ManualColorEncoding'] = 'ManualColorEncoding'
    array: ColorArray
    default: ColorValue = Field(default_factory=lambda: DEFAULT_COLOR)


class DirectColorEncoding(_DerivedStyleEncoding[ColorValue, ColorArray]):
    """Encodes color values directly from a feature column.

    Attributes
    ----------
    feature : str
        The name of the feature that contains the desired color values.
    fallback : ColorArray
        The safe constant fallback color to use if the feature column
        does not contain valid color values.
    """

    encoding_type: Literal['DirectColorEncoding'] = 'DirectColorEncoding'
    feature: str
    fallback: ColorValue = Field(default_factory=lambda: DEFAULT_COLOR)

    def __call__(self, features: Any) -> ColorArray:
        # A column-like may be a series or have an object dtype (e.g. color names),
        # neither of which transform_color handles, so convert to a list.
        return ColorArray.validate(list(features[self.feature]))


class NominalColorEncoding(_DerivedStyleEncoding[ColorValue, ColorArray]):
    """Encodes color values from a nominal feature whose values are mapped to colors.

    Attributes
    ----------
    feature : str
        The name of the feature that contains the nominal values to be mapped to colors.
    colormap : CategoricalColormap
        Maps the feature values to colors.
    fallback : ColorValue
        The safe constant fallback color to use if mapping the feature values to
        colors fails.
    """

    encoding_type: Literal['NominalColorEncoding'] = 'NominalColorEncoding'
    feature: str
    colormap: CategoricalColormap
    fallback: ColorValue = Field(default_factory=lambda: DEFAULT_COLOR)

    def __call__(self, features: Any) -> ColorArray:
        # map is not expecting some column-likes (e.g. pandas.Series), so ensure
        # this is a numpy array first.
        values = np.asarray(features[self.feature])
        return self.colormap.map(values)


class QuantitativeColorEncoding(_DerivedStyleEncoding[ColorValue, ColorArray]):
    """Encodes color values from a quantitative feature whose values are mapped to colors.

    Attributes
    ----------
    feature : str
        The name of the feature that contains the nominal values to be mapped to colors.
    colormap : Colormap
        Maps feature values to colors.
    contrast_limits : Optional[Tuple[float, float]]
        The (min, max) feature values that should respectively map to the first and last
        colors in the colormap. If None, then this will attempt to calculate these values
        from the feature values each time this generates color values. If that attempt
        fails, these are effectively (0, 1).
    fallback : ColorValue
        The safe constant fallback color to use if mapping the feature values to
        colors fails.
    """

    encoding_type: Literal['QuantitativeColorEncoding'] = (
        'QuantitativeColorEncoding'
    )
    feature: str
    colormap: Colormap
    contrast_limits: Optional[tuple[float, float]] = None
    fallback: ColorValue = Field(default_factory=lambda: DEFAULT_COLOR)

    def __call__(self, features: Any) -> ColorArray:
        values = features[self.feature]
        contrast_limits = self.contrast_limits or _calculate_contrast_limits(
            values
        )
        if contrast_limits is not None:
            values = np.interp(values, contrast_limits, (0, 1))
        return self.colormap.map(values)

    @validator('colormap', pre=True, always=True, allow_reuse=True)
    def _check_colormap(cls, colormap: ValidColormapArg) -> Colormap:
        return ensure_colormap(colormap)

    @validator('contrast_limits', pre=True, always=True, allow_reuse=True)
    def _check_contrast_limits(
        cls, contrast_limits
    ) -> Optional[tuple[float, float]]:
        if (contrast_limits is not None) and (
            contrast_limits[0] >= contrast_limits[1]
        ):
            raise ValueError(
                trans._(
                    'contrast_limits must be a strictly increasing pair of values',
                    deferred=True,
                )
            )
        return contrast_limits


def _calculate_contrast_limits(
    values: np.ndarray,
) -> Optional[tuple[float, float]]:
    contrast_limits = None
    if values.size > 0:
        min_value = np.min(values)
        max_value = np.max(values)
        # Use < instead of != to handle nans.
        if min_value < max_value:
            contrast_limits = (min_value, max_value)
    return contrast_limits
